"""
    adjoint - gradient based optimization 
"""
import os, shutil
import copy, random
from typing import Dict, List, Tuple
from scipy.optimize import minimize
import numpy as np
from tqdm import trange

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.optim import LBFGS, AdamW
from torch.utils.data import Dataset
from sklearn.model_selection import train_test_split
from ..helpers import MultiLayerLinear 
from ..helpers import diversity, distance, mutation_parameters, de_mutation_type, simple,de_dmp, de_best_1_bin,de_rand_1_bin
from ..helpers import uniform_reference_points, sort_and_select_population
from ..helpers import evaluation_func, transform_data, compute_mse
from ..base import Parameter, Individual, Optimizer
from .nsga3 import find_intercepts, find_extreme_points


class NSGA3_ML(Optimizer):
    def __init__(self,eval_command:str = "python evaluation.py", eval_folder:str = "Evaluation", optimization_folder:str=None,single_folder_eval:bool=False, overwrite_input_file:bool=False, linear_network:List[int]=[64,64,64,64],epochs:int=100, train_test_split:float=0.8,pop_size:int=32,ml_evals:int=5):
        """The objective of adjoint is to find the minimum of the jacobian of the evaluation parameters.

        Args:
            eval_command (str, optional): [description]. Defaults to "python evaluation.py".
            eval_folder (str, optional): [description]. Defaults to "Evaluation".
            optimization_folder (str, optional): Folder where the optimization and doe work should be stored in. Defaults to None.
            single_folder_eval (bool, optional): Evaluate within a single folder and not make a bunch of folders . Defaults to False.
            overwrite_input_file (bool, optional): whether or not to overwrite the input file with new data when restarting a simulation. Defaults to False.
            linear_network (list, optional): Size of MultiLinear network. Defaults to [128,256,256].
            epochs (int, optional): Number of epochs to train neural network for 
            train_test_split (float, optional): Number of datapoints to be assigned to train and test
            pop_size (int, optional): Population size 
            ml_evals(int, optional): Number of internal machine learning evaluations to perform before evaluating the next population

        """
        super().__init__(name="NSGA3_ML",eval_command=eval_command,eval_folder=eval_folder, opt_folder=optimization_folder,single_folder_eval=single_folder_eval,overwrite_input_file=overwrite_input_file)
        
        self.individuals = None
        self.linear_network = linear_network
        self.epochs = epochs
        self.train_test_split = train_test_split
        self.pop_size = pop_size
        self.__mutation_params = mutation_parameters()
        
        self.model = None
        self.optimizer = None
        self.label_scalers = None
        self.feature_scalers = None
        self.labels_str = None
        self.features_str = None
        self.ml_evals = ml_evals
    
    # * Mutation Properties
    @property
    def mutation_params(self):
        """Get Mutation parameters

        Returns:
            mutation_parameters: parameter class describes the mutation
        """
        return self.__mutation_params

    @mutation_params.setter
    def mutation_params(self,v:mutation_parameters):
        """Setter for mutation parameters 

        Args:
            v (mutation_parameters): class describing the mutation parameter s
        """
        self.__mutation_params = v

    def train(self,individuals:List[Individual],retrain:bool=False) -> Tuple[float, float]:
        """Trains the neural network to predict the output given an input 
    
        Optimizer (LBFGS): 
            https://johaupt.github.io/python/pytorch/neural%20network/optimization/pytorch_lbfgs.html 

        Args:
            individuals (List[Individual]): [description]
            retrain (bool, Optional): (True) retrains the existing model on new data. (False) create a new model

        Returns:
            Tuple[float float]: Train Loss and test loss 
        """
        
        # * Normalizing the Data
        if retrain: # If we simply train it with more data then there's no reason create new normalization scalers
            normalized_individuals, label_scalers, feature_scalers, labels_str, features_str = transform_data(individuals,self.label_scalers,self.feature_scalers)
        else:
            normalized_individuals, label_scalers, feature_scalers, labels_str, features_str = transform_data(individuals)
        self.label_scalers = label_scalers
        self.feature_scalers = feature_scalers
        self.labels_str = labels_str
        self.features_str = features_str

        labels = torch.as_tensor(np.array([ind.objectives for ind in normalized_individuals]),dtype=torch.float32)
        features = torch.as_tensor(np.array([ind.eval_parameters for ind in normalized_individuals]),dtype=torch.float32)
        
        # Transform
        data = list(zip(features,labels))
        test_size = int(len(data)*(1-self.train_test_split))
        train_size = int(len(data) - test_size)
        train_dataset, test_dataset = train_test_split(data,test_size=test_size,train_size=train_size,shuffle=True)
        train_dl = DataLoader(train_dataset,batch_size=128,shuffle=True)
        test_dl = DataLoader(test_dataset,batch_size=128,shuffle=False)
        # * Defining the Model        
        n_inputs = features.shape[1]
        n_outputs = labels.shape[1]

        if self.model is None or retrain == False:
            self.model = MultiLayerLinear(n_inputs,n_outputs,h_sizes=self.linear_network)
            # self.optimizer = LBFGS(self.model.parameters(), lr=0.0001,history_size=100, max_eval=int(20*1.25), max_iter=20)
            self.optimizer = AdamW(self.model.parameters())

        criterion = nn.MSELoss()
        
        for epoch in range(self.epochs):
            train_loss = 0 
            n_train = 0
            self.model.train()
            for i, (x, y) in enumerate(train_dl):
                batch_size = x.shape[0]
                self.optimizer.zero_grad()
                y_pred = self.model(x)
                loss = criterion(y_pred, y)
                loss.backward() # Zero gradients, backward pass, and update weights
                self.optimizer.step()
                # calculate the loss again for monitoring
                train_loss += loss.item()
                n_train += batch_size
            
            test_loss = 0
            n_test = 0
            self.model.eval()
            for j, (x,y) in enumerate(test_dl):
                batch_size = x.shape[0]
                y_pred = self.model(x)
                test_loss += criterion(y_pred, y).item()
                n_test += batch_size
            train_loss /= n_train
            test_loss /= n_test
        return train_loss, test_loss
            # print(f"Epoch: {epoch + 1:02}/{self.epochs} Train Loss: {train_loss:.5e} Test Loss: {test_loss:.5e}")
        
    def optimize_from_population(self,pop_start:int,n_generations:int):
        """Reads the values of a population, this can be a DOE or a previous evaluation
            Starts the optimization 

            Inputs:
                pop_start (-1 for DOE), reads the population folder and starts at pop_start+1
                n_generations - number of generations to run for
        """
        # * Read in all the results of the DOE, this should be done by a single thread
        # Check restart file, if not read the population
        self.load_history_file()

        newIndividuals = self.read_calculation_folder() # Use all individuals, there's no restart file
        newIndividuals = [item for sublist in newIndividuals for item in sublist]  # Flattens the list of lists
        # bounds = [(ep.min_value, ep.max_value) for ep in self.eval_parameters]

        if (len(newIndividuals)<self.pop_size):
            raise Exception("Number of individuals in the restart file is less than the population size."
                + " lower the population size or increase the DOE count(if restarting from a DOE)")

        # Do this before going into the train loop. This part of the code should happen after a new population is evaluated 
        ref_points = uniform_reference_points(len(self.objectives), p=4, scaling=None)
        individuals,best_point, worst_point, extreme_points = sort_and_select_population(individuals=newIndividuals,reference_points=ref_points, pop_size=self.pop_size)
        all_individuals = list() 
        all_individuals.extend(copy.deepcopy(individuals))

        for pop in range(pop_start+1,pop_start+n_generations):  # Population Loop 
            if self.ml_evals == 0:
                newIndividuals = self.__crossover_mutate__(individuals) # This becomes normal nsga3            
            ''' 
                Train a Neural network on Individuals. Initially this is all the individuals
            '''
            train_loss, test_loss = self.train(copy.deepcopy(all_individuals), False)
            
            '''
                Calculate new evaluation points using neural networks
            '''
            pop_dist_ml = list()
            pop_diversity_ml = list() 
            for _ in range(self.ml_evals): # perform nsga evaluations on the ML Model 
                newIndividuals = self.__crossover_mutate__(individuals)
                newIndividuals = evaluation_func(newIndividuals,self.model,self.label_scalers,self.feature_scalers)                
                pop_diversity_ml.append(diversity(newIndividuals))       # Calculate diversity 
                pop_dist_ml.append(distance(individuals,newIndividuals)) # Calculate population distance between past and future

                newIndividuals.extend(individuals) # add the previous population to the pool                                    

                individuals,best_point, worst_point, extreme_points = sort_and_select_population(newIndividuals,ref_points, self.pop_size) 

            # Evaluate
            self.evaluate_population(individuals,pop)
            # self.evaluate_population(newIndividuals,pop)
            newIndividuals = self.read_population(pop)
            mse = compute_mse(individuals,newIndividuals)
            print(f"mse {mse:03e}")
            # Sort and select
            pop_diversity = diversity(newIndividuals)       # Calculate diversity 
            pop_dist = distance(individuals,newIndividuals) # Calculate population distance between past and future
            all_individuals.extend(copy.deepcopy(newIndividuals))
            newIndividuals.extend(individuals) # add the previous population to the pool                                                
            individuals,best_point, worst_point, extreme_points = sort_and_select_population(newIndividuals,ref_points, self.pop_size)    # reduces the size of newIndividuals to the population size         
            self.append_restart_file(individuals)        # Keep the last designs
            self.append_history_file(pop,individuals[0],pop_diversity,pop_dist,train_loss,test_loss,mse)

            if self.single_folder_eval:
                # Delete the population folder
                population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(pop))
                if os.path.isdir(population_folder):
                    shutil.rmtree(population_folder)
  

    def add_eval_parameters(self,eval_params:List[Parameter]):
        """Add evaluation parameters. This is part of the initialization    

        Args:
            eval_params (List[Parameter]): Add in a list of evaluation parameters 

        """
        self.eval_parameters = eval_params # Sets base class variable

    def add_objectives(self,objectives:List[Parameter]):
        """Add the objectives 

        Args:
            objectives (List[Parameter]): [description]
        """
        self.objectives = objectives # Sets base class variable

    def add_performance_parameters(self,performance_params:List[Parameter] = None):
        """Add performance parameters 

        Args:
            performance_params (List[Parameter], optional): [description]. Defaults to None.
        """
        self.performance_parameters = performance_params # Sets base class variable
    
    def start_doe(self,doe_individuals:List[Individual]=None,doe_size:int=128):
        """Starts a design of experiments. This generates the parameters for the individuals to be evaluated and executes each case. If the DOE has already started and there is an output file for an individual then the individual won't be evaluated    

        Args:
            doe_individuals (List[Individual], optional): List of individuals. Defaults to None.
            doe_size (int, optional): [description]. Defaults to 128.
        """
        if doe_individuals is None:
            doe_individuals = []
            for i in trange(doe_size):
                parameters = copy.deepcopy(self.eval_parameters)
                for eval_param in parameters:
                    eval_param.value = np.random.uniform(eval_param.min_value,eval_param.max_value,1)[0]
                doe_individuals.append(Individual(eval_parameters=parameters,objectives=self.objectives, performance_parameters = self.performance_parameters))
     
        # * Begin the evaluation
        self.evaluate_population(individuals=doe_individuals,population_number=-1)
        # * Read the DOE
        individuals = self.read_population(population_number=-1)
        self.append_restart_file(individuals)
        
        if self.single_folder_eval: 
            # Delete the population folder
            population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(-1))
            if os.path.isdir(population_folder):
                shutil.rmtree(population_folder)


    def __crossover_mutate__(self,individuals:List[Individual]):
        """[summary]

        Args:
            individuals (List[Individual]): [description]

        Returns:
            [type]: [description]
        """

        nIndividuals = len(individuals)
        num_params = len(individuals[0].eval_parameters)        
        if self.mutation_params.mutation_type == de_mutation_type.de_best_1_bin:
            newIndividuals = de_best_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)
        elif self.mutation_params.mutation_type == de_mutation_type.de_rand_1_bin:
            newIndividuals = de_rand_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)
        elif self.mutation_params.mutation_type == de_mutation_type.de_dmp:
            newIndividuals = de_dmp(individuals=individuals,
                objectives=self.objectives, eval_parameters=self.eval_parameters, performance_parameters=self.performance_parameters)
        else:               # self.mutation_params.mutation_type == de_mutation_type.simple
            nCrossover =  int(self.pop_size/2)
            nMutation = self.pop_size-nCrossover
            newIndividuals = simple(individuals=individuals,nCrossover=nCrossover,nMutation=nMutation,objectives=self.objectives,eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,mu=self.mutation_params.mu,sigma=self.mutation_params.sigma)

        return newIndividuals
    


from dataclasses import dataclass, field, fields
from enum import Enum
from typing import List
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class Parameter:
    """This represents a value that a single evaluation keeps track of. For example the objective or if this is an computational simulation, we keep track of the stress, strain, volume, etc., anything that isn't a constraint. 

        Inputs:
            name: (string) name of the parameter 
            lower_value: (float) this represents the lower bound of the variable
            upper_value: (float) this represents the upper bound of the variable
            value_if_failed: (float) this is set to 100000, probably a good value for a minimization problem
            constr_less_than: (float) default is None for no constraint
            constr_greater_than: (float) default is None for no constraint
    """
    parameter = dict()
    name: str
    min_value:float = None
    max_value:float = None
    value_if_failed:float = 10000
    constraint_greater_than:float = None
    constraint_less_than:float = None
    value:float = 10000

class ParameterList:
    parameters:List[Parameter]



import sys
from typing import TypeVar,List
from .parameter import Parameter
import numpy as np

class Individual:
    """This class represents each individual or "evaluation". In each evaluation there are a set of parameters for objectives as well as additional parameters that are kept track. 
    """
    
    def __init__(self,eval_parameters:List[Parameter],objectives:List[Parameter],performance_parameters:List[Parameter]):
        """Initializes an individual with objectives and a list of parameters

        Args:
            eval_parameters (List[Parameter]): List of evaluation parameters. Evaluation parameters have a name and a min and max bound. This gives the optimizer a search space
            objectives (List[Parameter]): List of objectives. Each objective has a name and can be constrained 
            performance_parameters (List[Parameter]): List of performance parameters, stuff that you also want to track and may want to constrain, if you want.
        """
        self.__name = ""
        self.__population = -1
        self.__objectives = objectives
        self.__eval_parameters = eval_parameters
        self.__performance_parameters = performance_parameters
        
    
    def __hash__(self):        
        return hash(tuple(self.objectives))

    def __str__(self):        
        return str(tuple(self.objectives))
    
    def __repr__(self):
        """Returns a readable format when you are debugging using an IDE that is not jupyter notebook 

        Returns:
            str: readable format for debugging 
        """
        return "%s.(%s)(%s)" % (self.__class__.__name__,str(tuple(self.objectives)), str(tuple(self.eval_parameters)))

    @property 
    def name(self) -> str:
        """Name of the individual evaluation. For example IND000 

        Returns:
            str: name of the individual
        """
        return self.__name

    @name.setter
    def name(self,v:str):
        """Settor for name property

        Args:
            v (str): new name
        """
        self.__name = v

    @property 
    def population(self) -> int:
        """returns the population 

        Returns:
            int: [description]
        """
        return self.__population

    @population.setter
    def population(self,v:int):
        self.__population = v

    @property
    def objectives(self, include_constraint=True, C=0.5,a=2) -> np.ndarray:
        """Returns a numpy array of the objective values 

        Args:
            include_constraint (bool, optional): Whether or not to include constraints. Value of the objects are skewed if constraints are violated. Defaults to True.
            C (float, optional): Penalty coefficient. Defaults to 0.5.
            a (int, optional): Penalty exponential coefficient. Defaults to 2.

        Returns:
            np.ndarray: Array of values. For example [1,2]
        """
        y = np.zeros(len(self.__objectives))
        
        if (include_constraint): # apply constraint if constraint_value is negative
            y = self.__apply_dynamic_penalty(C=C,a=a)
        else:
            for i in range(len(self.__objectives)):
                y[i] = self.__objectives[i].value
            
        return y
    
    def constraints(self) -> np.ndarray:
        """Get the constraint matrix 

        Returns:
            np.ndarray: Return the constraint matrix 
        """
        perf_constraint = 0        
        if (self.performance_parameters is not None):
            for j in range(len(self.performance_parameters)):
                if (self.__performance_parameters[j].constraint_less_than is not None):
                    perf_constraint += (self.__performance_parameters[j].value - self.__performance_parameters[j].constraint_less_than)
                if (self.__performance_parameters[j].constraint_greater_than is not None):
                    perf_constraint += (self.__performance_parameters[j].constraint_greater_than - self.__performance_parameters[j].value)
        obj_constraint = 0
        for i in range(len(self.__objectives)):
            if (self.__objectives[i].constraint_less_than is not None):
                obj_constraint += (self.__objectives[i].constraint_less_than - self.__objectives[i].value)
            if (self.__objectives[i].constraint_greater_than is not None):
                obj_constraint += (self.__objectives[i].value - self.__objectives[i].constraint_greater_than)
        constraint = max([0,perf_constraint,obj_constraint])
        return constraint

    def __apply_dynamic_penalty(self,C:float,a:float) -> np.ndarray:
        """Applies dynamic penalty to the objectives 

        Args:
            C (float): penalty coefficient, this reduces the impact of bad designs at higher populations
            a (float): penalty exponential coefficient

        Returns:
            np.ndarray: objectives after the constraint has been applied 
        """
        y = np.zeros(len(self.__objectives))
        constraint = self.constraints()

        population = max([1,self.population]) # this way we avoid nan values
        for i in range(len(self.__objectives)):
            y[i] = self.__objectives[i].value + 1.0/(2.0*np.power(C*population,a)) * np.power(constraint,2) # Dynamic Penalty with Equation(2)
        return y
    
    @objectives.setter
    def objectives(self,v:List[Parameter]):
        """setter for objectives

        Args:
            v (List[Parameter]): array of parameters to set the objectives 
        """
        self.__objectives = v

    def get_objective(self,name:str) -> Parameter:
        """Searches for an objective based on a name provided and returns the objective as a parameter object

        Args:
            name (str): name of the objective. Example: "objective 1"

        Returns:
            Parameter: objective as parameter object 
        """
        temp = next((item.value for item in self.__objectives if item.name == name), None)
        return temp

    def get_objectives_list(self) -> List[Parameter]:
        """Returns a list of objectives as parameters 

        Returns:
            List[Parameter]: List of objectives 
        """        
        return self.__objectives
    
    def set_objective(self,name:str,val:float):
        """Assigns a value to a parameter for an individual
            typically called by the optimizer objectives[2] = 450
            
        Args:
            name (str): Name of the objective. Example: "Objective 1"
            val (float): value to set the objective to
        """
        for i in range(len(self.__objectives)):
            if (self.__objectives[i].name.lower() == name.lower()):
                self.__objectives[i].value = val     
                break
    
    @property
    def eval_parameters(self) -> np.ndarray:
        """Get eval parameters 

        Returns:
            np.ndarray: Returns a numpy array with the evaluation parameters 
        """
        y = np.zeros(len(self.__eval_parameters))
        for i in range(len(self.__eval_parameters)):
            y[i] = self.__eval_parameters[i].value
        return y
    
    @property
    def eval_parameter_min(self) -> np.ndarray:
        """Gets the minimum value of all the evaluation parameters

        Returns:
            np.ndarray: Array containing the minimum values of the evaluation paramters. 
        """
        y = np.zeros(len(self.__eval_parameters))
        for i in range(len(self.__eval_parameters)):
            y[i] = self.__eval_parameters[i].min_value
        return y
    
    @property
    def eval_parameter_max(self) -> np.ndarray:
        """returns the maximum value of each evaluation parameter as a numpy array 

        Returns:
            np.ndarray: Array containing maximum values of evaluation parameter
        """
        y = np.zeros(len(self.__eval_parameters))
        for i in range(len(self.__eval_parameters)):
            y[i] = self.__eval_parameters[i].max_value
        return y

    @eval_parameters.setter
    def eval_parameters(self,v):
        self.__eval_parameters = v
    
    def get_eval_parameter(self,name:str) -> Parameter:
        """Gets the evaluation parameter containing name

        Args:
            name (str): Name of the evaluation parameter to get

        Returns:
            Parameter: Parameter object containing name or None
        """        
        temp = next((item.value for item in self.__eval_parameters if item.name == name), None)
        return temp
    
    def set_eval_parameter_at_indx(self,indx:int,val:float):
        """Set the evaluation parameter value at a particular index

        Args:
            indx (int): Index in parameter array
            val (float): Value to replace with
        """
        self.__eval_parameters[indx].value = val

    def get_eval_parameter_list(self) -> List[Parameter]:
        """Gets all the evaluation parameters 

        Returns:
            List[Parameter]: List of evaluation parameters 
        """
        return self.__eval_parameters

    def set_eval_parameter(self,name:str,val:float):
        """set the evaluation parameter value given a name

        Args:
            name (str): Name of the parameter
            val (float): value to set to
        """
        for i in range(len(self.__eval_parameters)):
            if (self.__eval_parameters[i].name.lower() == name.lower()):
                self.__eval_parameters[i].value = val
    
    @property
    def performance_parameters(self) -> np.ndarray:
        """Gets the performance parameters as a numpy array

        Returns:
            np.ndarray: numpy array of performance parameters
        """
        if (self.__performance_parameters is not None):
            y = np.zeros(len(self.__performance_parameters))
            for i in range(len(self.__performance_parameters)):
                y[i] = self.__performance_parameters[i].value
            return y
        return None

    @performance_parameters.setter
    def performance_parameters(self,v:List[Parameter]):
        """Sets the performance parameters 

        Args:
            v (List[Parameter]): List of performance parameters defined as Parameter 
        """
        self.__performance_parameters = v
    
    def get_performance_parameter(self,name:str) -> Parameter:
        """Gets the performance parameter given a name

        Args:
            name (str): name of performance parameter to get 

        Returns:
            Parameter: Performance parameter or None
        """
        temp = next((item.value for item in self.__performance_parameters if item.name == name), None)
        return temp

    def set_performance_parameter_at_indx(self,indx:int,val:float):
        """Sets the performance parameter at an index 

        Args:
            indx (int): index of performance parameter 
            val (float): [description]
        """
        self.__eval_parameters[indx].value = val

    def set_performance_parameter(self,name:str,val:float):
        """Sets the performance parameter given the name

        Args:
            name (str): name of the parameter 
            val (float): value to set it to 
        """
        for i in range(len(self.__performance_parameters)):
            if (self.__performance_parameters[i].name.lower() == name.lower()):
                self.__performance_parameters[i].value = val

    def get_performance_parameters_list(self) -> List[Parameter]:
        """Gets a list of performance parameters 

        Returns:
            List[Parameter]: list of performance parameters 
        """
        return self.__performance_parameters

    @property
    def IsFailed(self) -> bool:
        """Checks the objectives to make sure the individual hasn't failed

        Returns:
            bool: True = failed, False = passed
        """
        objectives = [True for o in self.get_objectives_list() if o.value_if_failed == o.value]
        return any(objectives)  # Returns true if any of the values are True



#/bin/env python
"""
Optimizer - A base abstract class where all optimizers will inherit from. This class is never to be instantiated directly 
"""
"""
    Python Modules
"""
from os import name
import sys
import os, glob, copy, signal, platform, ctypes
from typing import TypeVar,List, Dict, Tuple
import subprocess
import time
import math
import pprint
import logging
from warnings import catch_warnings
import pandas as pd
"""
    External Modules
"""
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
from .individual import Individual
from .parameter import Parameter
from ..helpers import copy_helper, parallel_settings, convert_to_ndarray, check_if_duplicates
import psutil

class Optimizer: 
    """
        Base class for starting an optimization
    """
    def __init__(self,name:str,eval_command:str, eval_folder:str = None,opt_folder:str=None, eval_parameters:List[Parameter]=[], objectives:List[Parameter] = [], performance_parameters:List[Parameter] = [], single_folder_eval=False, overwrite_input_file=False):
        """Initializes the optimizer. This is typically inherited by a particular type of optimizer like NSGA (multi-objective differential evolution) or SODE (single objective differential evolution). Any new optimization strategy should inherit from this class.
        This class handles most of the heavy work of evaluation and reading the results.

        Args:
            name (str): name of the optimizer
            eval_command (str): This is the location of the evaluation script. Optimizer will call this script and read output.txt containing the values of the objectives and any performance parameters, see examples. 
            eval_folder (str, optional): If specified, the evaluation folder will be copied for each execution. Defaults to None.
            opt_folder (str, optional): Working directory where you want to start the optimization (calculatio folder will be created here). Defaults to None.
            eval_parameters (List[Parameter], optional): This is the list of x1-xn that feeds into your execution code. Defaults to [].
            objectives (List[Parameter], optional): The objectives you want to keep track of and the value for when they fail. Defaults to [].
            performance_parameters (List[Parameter], optional): Number of parameters. Defaults to [].
            single_folder_eval (bool, optional): Saves space by deleting the population folder when completed. by default this is false. Defaults to False.
            overwrite_input_file(bool, optional): Specifies whether or not to overwrite the input file when restarting a simulation. Defaults to False.

        Raises:
            Exception: Invalid relative path could not be found for eval script
            Exception: Invalid relative path could not be found for eval folder
            Exception: Invalid absolute path could not be found for eval script
            Exception: Invalid absolute path could not be found for eval folder
        """
        self.name = name        
        assert opt_folder is not None

        try:
            logging.basicConfig(filename=os.path.join(opt_folder,'log.dat'),  level=logging.DEBUG)
            self.logger = logging.getLogger()
        except:
            self.logger = None
        

        if (not os.path.isabs(eval_folder)):
            eval_folder = os.path.join(os.getcwd(),eval_folder)
            if (not os.path.isdir(eval_folder)):
                raise Exception('Invalid relative path could not be found for eval folder {0}'.format(eval_folder))

        if (not os.path.isdir(eval_folder)):
            raise Exception('Invalid absolute path could not be found for eval folder {0}'.format(eval_folder))


        self.evaluation_folder = eval_folder
        self.evaluation_command = eval_command

        if (opt_folder):
            self.optimization_folder = opt_folder
        else:
            self.optimization_folder = os.getcwd()        
        
        if (not os.path.isabs(self.optimization_folder)):
            self.optimization_folder = os.getcwd()


        self.population_track = []
        self.overwrite_input_file = overwrite_input_file
        self.input_file = 'input.dat'
        self.output_file = 'output.txt'
        self.__use_calculation_folders = True

        # Optimization variables to track
        self.eval_parameters = eval_parameters
        self.objectives = objectives
        self.performance_parameters = performance_parameters
        self.__parallel_settings = parallel_settings()
        self.cores_per_evalulation = []
        # Cache storage
        self.pandas_cache = {} # Appends individuals to pandas dataframe each dictionar contains the population number
        self.single_folder_eval = single_folder_eval

        
        objective_names = [o.name for o in self.objectives]
        eval_parameter_names = [o.name for o in self.eval_parameters]
        performance_parameter_names = [o.name for o in self.performance_parameters]
        assert check_if_duplicates(objective_names) == False, "Objective names have to be unique"
        assert check_if_duplicates(eval_parameter_names) == False, "Evaluation Parameter names have to be unique"
        assert check_if_duplicates(performance_parameter_names) == False, "Performance Parameter names have to be unique"

        
    @property
    def use_calculation_folder(self) -> bool:
        """Allows the optimizer to define calculation folders for each call.

        Returns:
            bool: True = use calcuation folder  
        """
        return self.__use_calculation_folder

    @use_calculation_folder.setter
    def use_calculation_folder(self,b:bool=True):
        """ Allows the optimizer to define calculation folders for each call.

        Args:
            b (bool): True = use calcuation folder. Defaults to True
        """
        self.__use_calculation_folder=b

    def change_working_dir(self,new_dir:str):
        """Changes the current working directory

        Args:
            new_dir (str): path to new directory 
        """
        self.optimization_folder = new_dir
    
    def get_current_directory(self) -> str:
        """returns the current working directory

        Returns:
            str: Path to current directory
        """
        return self.optimization_folder
    
    @property
    def parallel_settings(self):
        """Returns the parallel settings dataclass 

        Returns:
            parallel_settings: parallel settings class object
        """
        return self.__parallel_settings

    @parallel_settings.setter
    def parallel_settings(self,settings:parallel_settings):
        """Sets the parallel settings
            This also reads the master machine file if exists and splits up the cores per execution.
            Machine file is basically the computer name followed by an index. See an example below:
                paht-ryzen0
                paht-ryzen1
                paht-ryzen2
                paht-ryzen3
                steve-intel0
                steve-intel1
                steve-intel2
                steve-intel3
                
        Args:
            settings (parallel_settings): dataclass representing the parallel settings to use
        """
        self.__parallel_settings = settings
        # Check for machine file
        machinefile = os.path.join(self.optimization_folder,settings.machine_filename)        
        if os.path.exists(machinefile):
            with open(machinefile, 'r') as f:
                x = f.read().splitlines()
                temp = []
                for i in range(len(x)):
                    temp.append(x[i])
                    if self.__parallel_settings.cores_per_execution > 0:
                        if (i+1) % self.__parallel_settings.cores_per_execution == 0:
                            self.cores_per_evalulation.append(copy.deepcopy(temp))   # Splits up the machine file
                            temp.clear()
                    

    def __create_input_file__(self,individual:Individual):
        """Creates an input file 'inputs.txt' which the evaluation script reads. Will only  create if inputs.txt does not already exist

        Args:
            individual (Individual): Individual's evaluation parameters are read x[1-N] are printed to an evaluation script
        """
        def write_input():
            with open(self.input_file,'w') as f:
                eval_params = individual.get_eval_parameter_list()
                for p in range(len(eval_params)):
                    f.write('{0} = {1}\n'.format(eval_params[p].name,eval_params[p].value))

        if os.path.exists(self.input_file) and self.overwrite_input_file:
            write_input()
        else:       # input file does not exists
            write_input()
       
    def __read_output_file__(self, individual:Individual) -> Individual:
        """Reads the output file i.e. output.txt line by line and parses it for objectives and parameters

        Args:
            individual (Individual): Individual object that is read and parameters are set 

        Returns:
            Individual: [description]
        """
        # Read the output file
        if (os.path.exists(self.output_file)):
            with open(self.output_file,'r') as f:
                for line in f:
                    split_val = line.split('=')
                    key = split_val[0].strip()
                    val = float(split_val[1].strip())
                    individual.set_objective(name=key,val=val)
                    individual.set_performance_parameter(name=key,val=val)
        else:   # no output.txt found, something bad happened
            for objective in individual.get_objectives_list():                
                individual.set_objective(name=objective.name,val=objective.value_if_failed)
            for param in individual.get_performance_parameters_list():
                individual.set_performance_parameter(name=param.name,val=param.value_if_failed)
        return individual
    
    def __read_input_file__(self, individual:Individual) -> Individual:
        """Reads the input file i.e. input.dat line by line and parses it for evaluation parameters

        Args:
            individual (Individual): Individual object that will be set

        Returns:
            Individual: returns an individual with updated values 
        """
        # Read the input file
        with open(self.input_file,'r') as f:
            for line in f:
                split_val = line.split('=')
                key = split_val[0].strip()
                val = float(split_val[1].strip())
                individual.set_eval_parameter(name=key,val=val)
        return individual
    
    def __check_population_folder__(self,population_number:int=0) -> str:
        """Formats the population folder

        Args:
            population_number (int, optional): Population Number for example 20. Defaults to 0.

        Returns:
            str: relative path to the population folder 
        """

        if (population_number<0): #  Check current directory for calculation folder
            population_folder = "Calculation/DOE"
        else:
            population_folder = "Calculation/POP{:03d}".format(population_number)
        return population_folder


    def read_population(self,population_number:int=0) -> List[Individual]:
        """Reads the output file in the population/DOE folder

        Args:
            population_number (int, optional): Which population we are evaluating. Defaults to 0.

        Raises:
            Exception: Restart population was not found

        Returns:
            List[Individual]: list of individuals within the population
        """
        population_folder = self.__check_population_folder__(population_number)
        pop_path = os.path.join(self.optimization_folder,population_folder)
        if (not os.path.exists(pop_path)):
            raise Exception("Restart population was not found")
        else:
            ind_directories = [dI for dI in os.listdir(pop_path) if os.path.isdir(os.path.join(pop_path,dI)) and ('IND' in dI)]
        
        ind_directories = sorted(ind_directories)
        
        individuals = list()
        for ind_dir in ind_directories:
            ind = Individual(objectives=copy.deepcopy(self.objectives),eval_parameters=copy.deepcopy(self.eval_parameters),performance_parameters=copy.deepcopy(self.performance_parameters))
            ind.name = ind_dir
            ind.population = population_number
            current_directory = os.getcwd()
            ind_dir = os.path.join(pop_path,ind_dir)
            os.chdir(ind_dir)
            ind = self.__read_output_file__(ind)
            ind = self.__read_input_file__(ind)
            individuals.append(copy.deepcopy(ind))
            os.chdir(current_directory)

        return individuals

    def __select_cores_per_execution__(self,pid_list:list):
        """Returns the index of cores_per_execution that is not being used 
            Used to find out which process id has been freed 

        Args:
            pid_list (list): list of process ids from pOpen

        Returns:
            int: index of process id that is no longer being used. 
        """

        if len(pid_list)==0:
            return 0
        c_range = range(len(self.cores_per_evalulation))
        c_range_used = [p['cores_per_execution_indx'] for p in pid_list]
        available_indicies = [x for x in c_range if x not in c_range_used]
        if (len(available_indicies)>0):
            return available_indicies[0]             

    def evaluate_population(self,individuals:List[Individual],population_number:int=0):
        """Evaluates a population of individuals, checks to see if the population exists in the directory. If it already exists, do nothing, read the results, evaluate if there is no output file

        Args:
            individuals (List[Individual]): List of individuals to evaluate 
            population_number (int, optional): population corresponding to the set of individuals. Defaults to 0.
        """
        
        population_folder = self.__check_population_folder__(population_number)

        pop_path = os.path.join(self.optimization_folder,population_folder)
        if (not os.path.exists(pop_path)):
            os.makedirs(pop_path)
        
        if len(self.cores_per_evalulation)>0:     # Determine whether to use machine file or not
            n_evalulations = len(self.cores_per_evalulation)
            use_cores = True
        else:
            n_evalulations = self.parallel_settings.concurrent_executions
            use_cores = False

        pid_list = list()   # keep track of the process id and start time        
        eval_count = 0;# c is the counter for cores per interation [['paht_cpu1','paht_cpu1','paht_cpu1'],['paht_cpu2','paht_cpu2','paht_cpu2']]
        # Evaluate each individual
        for ind_indx in range(len(individuals)):            
            individuals[ind_indx].name = 'IND{:03d}'.format(ind_indx)
            individuals[ind_indx].population = population_number
            ind_dir = os.path.join(pop_path, individuals[ind_indx].name)
            if (not os.path.exists(ind_dir)):
                os.makedirs(ind_dir)
            # Check if eval folder exists
            if (self.evaluation_folder):
                # Evaluate Individual (calls method in base class)
                if (use_cores):
                    c = self.__select_cores_per_execution__(pid_list)
                    pid,p = self.__evaluate_individual__(individual=individuals[ind_indx],individual_directory=ind_dir,cores_per_execution=self.cores_per_evalulation[c])
                    pid_list.append({'pid':pid,'start_time':time.perf_counter(),'cores_per_execution_indx':c,'proc':p,'pop':str(individuals[ind_indx].population),'ind':individuals[ind_indx].name})
                else:                    
                    pid,p = self.__evaluate_individual__(individual=individuals[ind_indx],individual_directory=ind_dir)
                    pid_list.append({'pid':pid,'start_time':time.perf_counter(),'proc':p,'pop':str(individuals[ind_indx].population),'ind':individuals[ind_indx].name})
                
                while (len(pid_list)==n_evalulations): # Pause any new executions once we reach maximum number of evaluations 
                    inActive_pids = list()
                    # Loop to check if PID is still active and execution time is less than timeout                    
                    for i in range(len(pid_list)):
                        pid = pid_list[i]['pid']; p = pid_list[i]['proc']; pop = pid_list[i]['pop']; ind_name = pid_list[i]['ind']
                        if self.__check_process_running__(p): # If PID is running check if it exceeded the execution time
                            start_time = pid_list[i]['start_time']
                            time.sleep(0.1) # Check every "x" seconds
                            if (time.perf_counter() - start_time)/60 > self.parallel_settings.execution_timeout:
                                try:
                                    self.__write_proc_log__(p,pop,ind_name)
                                    os.kill(pid, signal.SIGTERM)
                                except:
                                    pass
                                inActive_pids.append(i)
                        else: # PID isn't running, remove it
                            self.__write_proc_log__(p,pop,ind_name)
                            inActive_pids.append(i)
                    
                    for index in sorted(inActive_pids, reverse=True): # removing the inactive PIDs from the list 
                        del pid_list[index]  
                    
            else: # TODO execute individual without creating a bunch of directories. Maybe create execution directories and delete them
                output = subprocess.check_output([self.evaluation_command])
                # Append output to results, need to check first how the output is structured
        time.sleep(0.1)
    
    def __check_process_running__(self,p:subprocess.Popen) -> bool:
        """Checks if a process id is running. This checks by using the poll method.

        Args:
            p (subprocess.Popen): Popen class 

        Returns:
            bool: true if process id is still running 
        """
        if p is not None:
            poll = p.poll()
            if poll == None:
                return True
        return False

    def __check_PID_running__(self,pid:int) -> bool:
        """Checks if process id is running. this checks by trying to kill the process. if there's an error then it's running

        Args:
            pid (int): integer corresponding to the process id 

        Returns:
            bool: true if process is still running
        """
        if (platform.system() == 'Linux'):
            try:
                os.kill(pid, 0)
                if pid<0:               # In case code terminates
                    return False
            except OSError:
                return False 
            else:
                return True
        elif (platform.system() == 'Windows'):
            return pid in (p.pid for p in psutil.process_iter())
            
    
    
    def __evaluate_individual__(self,individual:Individual,individual_directory:str,cores_per_execution:list=[]) -> Tuple[int,subprocess.Popen]:
        """Evaluates the individual by copying the evaluation folder into the individual's directory

        Args:
            individual (Individual): [description]
            individual_directory (str): [description]
            cores_per_execution (list, optional): [description]. Defaults to [].

        Returns:
            (tuple): tuple containing:

                **pid** (int): process id
                **process** (subprocess.Popen): logging message string 
        """
        # Check for output.txt
        if (not os.path.exists(os.path.join(individual_directory,self.output_file))): # This prevents an evaluation if there is already an output
            copy_helper.copy(self.evaluation_folder,individual_directory)
            cur_dir = os.getcwd()
            os.chdir(individual_directory)            
            # Create the input file
            self.__create_input_file__(individual)
            # Create local machine file            
            if (len(cores_per_execution)>0):                
                with open(self.parallel_settings.machine_filename,'w') as f:
                    f.write('\n'.join(cores_per_execution))
                    
            # Evaluate
            p = subprocess.Popen(self.evaluation_command.split(' '),stdout=subprocess.PIPE, stderr=subprocess.STDOUT) # Note: evaluation_command has to read the local machine file. This is not my problem            
            os.chdir(cur_dir)
            return p.pid, p
        return -1,None

    def __write_proc_log__(self,p:subprocess.Popen,pop:int,ind_name:str):
        """Write the log for each process 

        Args:
            p (subprocess.Popen): the Popen object 
            pop (int): population as a ninteger 
            ind_name (str): name of the individual 
        """
        if p is not None:     
            for line in p.stdout:
                if (self.logger):
                    self.logger.debug('POP {0} Indivudual: {1} Message: {2}'.format(pop,ind_name,line.decode("utf-8").replace('\n', ' ').replace('\r', '')).strip())
        
    def load_history_file(self):
        """Reads the history file if exists 
        """
        self.__history_filename = os.path.join(self.optimization_folder,"history.csv")
        if (os.path.exists(self.__history_filename)):
            df_temp = pd.read_csv(self.__history_filename)
            # Check for matching columns
            headers = ['Unnamed: 0','Population','Best Individual']
            eval_param_names =  [p.name for p in self.eval_parameters]
            objective_names = [o.name for o in self.objectives]
            perf_param_names = [o.name for o in self.performance_parameters]
            
            headers.extend(eval_param_names)
            headers.extend(objective_names)
            headers.extend(perf_param_names)
            headers.extend(['pop_diversity','pop_avg_distance'])
            

            if len(headers) == len(list(df_temp.columns)) and headers == list(df_temp.columns):
                self.history = df_temp
            else:   # if the number of parameters change then start from a new file 
                self.history = None
                os.remove(self.__history_filename)

    def append_history_file(self, pop:int, best_ind:Individual,diversity:float,distance:float, train_loss:float=0, test_loss:float=0,mse:float=0):
        """Writes a history.csv file containing the best design(s) this function is called by the inheriting class

        Args:
            pop (int): Population index
            best_ind (Individual): best performing individual 
            diversity (float): diversity of the population 
            distance (float): average distance between individuals of the population
        """
        eval_params = best_ind.eval_parameters
        eval_param_names =  [p.name for p in best_ind.get_eval_parameter_list()]
        
        objectives = best_ind.objectives
        objective_names = [o.name for o in best_ind.get_objectives_list()]
        
        perf_param = best_ind.performance_parameters
        perf_param_names = [o.name for o in best_ind.get_performance_parameters_list()]
        
        header = ['Population', 'Best Individual']
        pop_dir = self.__check_population_folder__(best_ind.population).replace('Calculation/','')
        data = [pop, '{:s}_{:s}'.format(pop_dir,best_ind.name)]
        
        def write_arrays(names,vals):
            for name,val in zip(names,vals):
                header.append(name)
                data.append(val)

        write_arrays(eval_param_names,eval_params)
        write_arrays(objective_names,objectives)
        write_arrays(perf_param_names,perf_param)
        
        header.extend(['pop_diversity','pop_avg_distance','train_loss','test_loss','actual_mse_loss'])
        data.extend([diversity,distance,train_loss,test_loss,mse])
        if (not os.path.exists(self.__history_filename)):                        
            self.history = pd.DataFrame(dict(zip(header, data)),index=[0])
            self.history.to_csv(self.__history_filename)
        else:            
            self.history = pd.concat([self.history,pd.DataFrame(dict(zip(header, data)),index=[1])])
            self.history.to_csv(self.__history_filename)

    def append_restart_file(self, individuals:List[Individual]):
        """Appends self.population_track to a restart file, these are the individuals matter and can be restarted from. Instead of restarting from a population, lets restart from the best individuals 

        Args:
            individuals (List[Individual]): list of individuals, this can be any size. these individuals will be added to the restart file
        """
        df = self.to_pandas(individuals=individuals,bReturnPandas=True)
        df.to_csv(os.path.join(self.optimization_folder,'restart_file.csv'))

   
    def read_restart_file(self) -> List[Individual]:
        """Appends self.population_track to a restart file, these are the individuals matter and can be restarted from. Instead of restarting from a population, lets restart from the best individuals 

        Raises:
            Exception: Error about restart file not being the same. this can happen if you add evaluation parameters or performance parameters. Make sure number of columns match the number of parameters. You can add null or dummy values. 

        Returns:
            List[Individual]: individual list or empty list
        """

        restart_file = os.path.join(self.optimization_folder,'restart_file.csv')
        if os.path.exists(restart_file):
            df = pd.read_csv(restart_file)
            df_columns = df.columns
            n_eval = len(self.eval_parameters)
            n_obj = len(self.objectives)
            n_perf = len(self.performance_parameters)
            if ((len(df_columns)-3) != (n_eval+n_obj+n_perf)):
                raise Exception("this is not the right restart file. number of columns does not match the sum of evaluation, objective, and performance parameters")
            else:
                individual_list = []
                offset = 3 # this is the column where data starts
                for index,row in df.iterrows():
                    ind = Individual(eval_parameters=self.eval_parameters,objectives=self.objectives,performance_parameters=self.performance_parameters)
                    ind.name = row['individual']
                    ind.population = int(row['population'].replace('DOE','-1').replace('POP',''))
                    [ind.set_eval_parameter(p.name,row[p.name]) for p in self.eval_parameters]
                    [ind.set_objective(p.name,row[p.name]) for p in self.objectives]
                    [ind.set_performance_parameter(p.name,row[p.name]) for p in self.performance_parameters]                    
                    individual_list.append(copy.deepcopy(ind))
                return individual_list
        
        return []

    def to_pandas(self,individuals:List[Parameter],pop_number:int=0,bReturnPandas:bool = False):
        """Exports the results to a pandas file 

        Args:
            individuals (List[Parameter]): [description]
            pop_number (int, optional): [description]. Defaults to 0.
            bReturnPandas (bool, optional): [description]. Defaults to False.

        Returns:
            pd.DataFrame: dataframe object with all the optimization results 
        """

        def add_value_to_dict(dict_obj,key,val):
            indx = 1
            while (key in dict_obj):
                key = key + "_" + str(indx)
                indx+=1
            dict_obj[key] = val
            return dict_obj
        
        # Create data as a dict of names, the only thing that will break this is if the dictionary has repeated keys
        data = []
        for individual in individuals:
            if (individual.population<0):
                Population_Name = 'DOE'
            else:
                Population_Name = 'POP{0:03d}'.format(individual.population)
            
            temp_dict = {}
            add_value_to_dict(temp_dict,'population',Population_Name) # name of the population IND000 for example
            add_value_to_dict(temp_dict,'individual',individual.name) # name of the individual IND000 for example
            
            for eval_param in individual.get_eval_parameter_list():            
                add_value_to_dict(temp_dict,eval_param.name,eval_param.value)

            for objective in individual.get_objectives_list():
                add_value_to_dict(temp_dict,objective.name,objective.value)

            for perf_param in individual.get_performance_parameters_list():
                add_value_to_dict(temp_dict,perf_param.name,perf_param.value)
                
            data.append(temp_dict) # TODO Check if this is duplicated, if so then probably needs to be copied
        if (bReturnPandas):
            return pd.DataFrame(data)

        if pop_number<0:
            self.pandas_cache['DOE'] = pd.DataFrame(data)
        else:
            self.pandas_cache['POP{0:03d}'.format(pop_number)] = pd.DataFrame(data)

    @staticmethod
    def df_to_tecplot(df_dict:Dict[str,pd.DataFrame],filename:str):
        """Staticmethod to convert a normal pandas dataframe to tecplot file

        Args:
            df (Dict[str,pd.DataFrame]): For example {'DOE': pd.DataFrame}
        """
        firstKey = next(iter(df_dict))
        firstPandas = df_dict[firstKey]

        # header
        variables = []
        for col in firstPandas.columns:
            variables.append('\"{0}\"'.format(col))
        head = 'VARIABLES = ' + ','.join(variables) + '\n'

        # zones
        zones = []
        for key, df in df_dict.items():
            i = 0
            for index, row in df_dict[key].iterrows():
                zone_str = 'ZONE T = \"{0}\"\n'.format(key+"_"+df.iloc[i]['individual'])
                data_str = ' '
                data = []
                i+=1
                for col in df_dict[key].columns:
                    if col=='population':
                        if row[col] == "DOE":
                            data.append(str(-1))
                        else:
                            data.append(str(row[col].replace('POP','')))
                    elif col=="individual":
                        data.append(str(row[col].replace('IND','')))
                    else:
                        if type(row[col])== float:
                            data.append("{:.6E}".format(row[col]))
                        elif type(row[col])== int:
                            data.append("{0}".format(row[col]))
                        else:
                            data.append(row[col])
                zones.append(zone_str)
                zones.append(' '.join(data) + '\n')
        
        # write to .tec file
        with open(filename,'w') as f:
            # write headers
            f.write(head)
            # write zones
            for zone in zones:
                f.write(zone)

    def to_tecplot(self,use_min_max:bool=False):
        """Converts the dataframe to a tecplot file (.tec)
        """
        if len(self.pandas_cache)==0:
            return
        if (not os.path.exists(os.path.join(self.optimization_folder,''))):
            os.makedirs(os.path.join(self.optimization_folder,''))

        filename = os.path.join(self.optimization_folder,'','database.tec')
        for i in range(len(self.objectives)):
            if use_min_max == True:
                if self.objectives[i].max_value is not None:
                    column = self.objectives[i].name
                    max_value = self.objectives[i].max_value
                    for k in self.pandas_cache.keys():
                        self.pandas_cache[k] = self.pandas_cache[k][(self.pandas_cache[k][column] <= max_value)]
                        
                if self.objectives[i].min_value is not None:
                    column = self.objectives[i].name
                    min_value = self.objectives[i].min_value
                    for k in self.pandas_cache.keys():
                        self.pandas_cache[k] = self.pandas_cache[k][(self.pandas_cache[k][column] >= min_value)]
        self.df_to_tecplot(self.pandas_cache,filename)

    
    def create_restart(self):
        """Create a restart file containing all individuals of all populations
        """
    
        pop_individuals = self.read_calculation_folder()
        for individuals in pop_individuals:
            self.append_restart_file(individuals)

    def plot_2D(self,obj1_name:str,obj2_name:str,xlim:list=None,ylim:list=None):
        """Creates a 2D plot scatter plot of all the individuals for the two objectives specified

        Args:
            obj1_name (str): name of the x-axis
            obj2_name (str): name of the y-axis
            xlim (list, optional): xbounds example [-1,2]. Defaults to None.
            ylim (list, optional): ybounds example [-5,5]. Defaults to None.
        """
        fig,ax = plt.subplots()

        colors = cm.rainbow(np.linspace(0, 1, len(self.pandas_cache.keys())))        
        indx = 0
        legend_labels = []
        # Scan the pandas file, grab objectives for each population
        for key, df in self.pandas_cache.items():
            obj1_data = []
            obj2_data = []
            c=colors[indx]
            for index, row in df.iterrows():
                obj1_data.append(row[obj1_name])
                obj2_data.append(row[obj2_name])
            # Plot the gathered data
            ax.scatter(obj1_data, obj2_data, color=c, s=5,alpha=0.5)
            legend_labels.append(key)
            indx+=1

        ax.set_xlabel(obj1_name)
        ax.set_ylabel(obj2_name)
        if xlim is not None:
            ax.set_xlim(xlim[0],xlim[1])
        if ylim is not None:
            ax.set_ylim(ylim[0],ylim[1])
        ax.legend(legend_labels)
        fig.canvas.draw()
        fig.canvas.flush_events()
        plt.show()
 
    def read_calculation_folder(self) -> List[Individual]:
        """Reads the entire calculation folder to a dataframe and returns all the individuals as an array
            this can be useful for restarting a population

        Raises:
            Exception: [description]

        Returns:
            List[Individual]: list of all individuals for all populations (could be used as a restart)
        """
        
        pop_path = os.path.join(self.optimization_folder,'Calculation')
        try:
            list_subfolders = [int(f.name.replace('POP','').replace('DOE','-1')) for f in os.scandir(pop_path) if f.is_dir()]
        except:
            raise Exception("Invalid directory found. Make sure directories are either DOE or POP001")
        
        individual_list = []
        self.pandas_cache = dict()
        for pop_indx in list_subfolders:
            individuals = self.read_population(pop_indx)            
            self.to_pandas(copy.deepcopy(individuals),pop_indx)
            individual_list.append(copy.deepcopy(individuals))
        return individual_list


    def to_dict(self):
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            dict: list of all the settings, could be used for creating an optimization object
        """
        settings = dict()
        settings['eval_folder'] = self.evaluation_folder
        settings['opt_folder'] = self.optimization_folder
        settings['single_folder_eval'] = self.single_folder_eval
        settings['eval_parameters'] = [p.to_dict() for p in self.eval_parameters]
        settings['objectives'] = [o.to_dict() for o in self.objectives]
        settings['perf_parameters'] = [p.to_dict() for p in self.performance_parameters]

        self.read_calculation_folder()
        results = dict()
        for k in self.pandas_cache.keys():
            results[k] = self.pandas_cache[k].to_dict()
        
        settings['results'] = results
        return settings
    
    def from_dict(self,settings:dict):
        """Reads the dictionary file and creates the base object

        Args:
            settings (dict): dictionary created by to_dict()
        """
        self.evaluation_folder = settings['eval_folder']
        self.optimization_folder = settings['opt_folder']
        self.single_folder_eval = settings['single_folder_eval']
        
        self.eval_parameters = [Parameter().from_dict(p) for p in settings['eval_parameters']]
        self.objectives = [Parameter().from_dict(p) for p in settings['objectives']]
        self.performance_parameters = [Parameter().from_dict(p) for p in settings['perf_parameters']]
        self.read_calculation_folder()


from typing import Dict, List, Tuple
import torch 
import torch.nn as nn
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from ..base import Individual
import numpy as np 
from sklearn.metrics import mean_squared_error

@torch.no_grad()
def evaluation_func(individuals:List[Individual], model:nn.Module, label_scalers:Dict[str,MinMaxScaler], feature_scalers:Dict[str,MinMaxScaler]):
    """Evaluates the individuals with the neural network function 

    Args:
        x0 (np.ndarray): [description]
        model (nn.Module): [description]
        reference_points (List[np.ndarray]): [description]
        intercepts (np.ndarray)
    """
    individuals, _, _, labels_str, features_str = transform_data(individuals, label_scalers, feature_scalers)    

    features = np.array([ind.eval_parameters for ind in individuals])
    features = torch.as_tensor(features,dtype=torch.float32)
    fitnesses = model(features)

    for i in range(len(individuals)):
        for j,l in enumerate(labels_str):
            value = label_scalers[l].inverse_transform(fitnesses[i,j].detach().numpy().reshape(1,-1))
            individuals[i].set_objective(l, value[0,0])

        for k,l in enumerate(features_str):
            value = feature_scalers[l].inverse_transform(features[i,k].detach().numpy().reshape(1,-1))
            individuals[i].set_eval_parameter(l, value[0,0])

    return individuals

   
def transform_data(individuals:List[Individual],label_scalers:Dict[str,MinMaxScaler]=None, feature_scalers:Dict[str,MinMaxScaler]=None) -> List[Individual]:
    """Normalizes all the individuals and returns the label scaler for objectives and feature scalers for the evaluation parameters 

    Args:
        individuals (List[Individual]): Unnormalized list of individuals 
        label_scalers (Dict[str,MinMaxScaler]): scalers that can be used to untransform the objectives. Optional, Default = None
        feature_scalers (Dict[str,MinMaxScaler]): scalers that can be used to untransform the evaluation parameters. Optional, Default = None

    Returns:
        (tuple): Tuple containing

            **Individuals** (List[Individual]): Normalized list of individuals
            **label_scalers** (Dict[str,MinMaxScaler]): scalers that can be used to untransform the objectives
            **feature_scalers** (Dict[str,MinMaxScaler]): scalers that can be used to untransform the evaluation parameters

    """
    individuals = [ind for ind in individuals if ind.IsFailed==False] # remove failed simulations

    labels = np.array([ind.objectives for ind in individuals])
    labels_str = [o.name for o in individuals[0].get_objectives_list()]

    features = np.array([ind.eval_parameters for ind in individuals])
    eval_parameters = individuals[0].get_eval_parameter_list()
    feature_bounds = [(ep.min_value, ep.max_value) for ep in eval_parameters]
    features_str = [f.name for f in individuals[0].get_eval_parameter_list()]

    def get_scalers(data:np.ndarray,data_keys:List[str], bounds:List[Tuple[float,float]]=None):
        """This returns the scaled data

        Args:
            data (np.ndarray): data as a numpy array
            data_keys (List[str]): [description]
            bounds (List[Tuple[float,float]], optional): If you specify a list of min and max [(0.1, 0.5)], we use these ranges to scale your features. Defaults to None.

        Returns:
            (tuple): containing

                **scalers** (MinMaxScaler): minmax scaler object
                **new_data** (np.ndarray): scaled data
        """
        new_data = np.zeros(data.shape)
        scalers = dict() # Scale each label
        for i in range(len(data_keys)):
            label_name = data_keys[i]
            scalers[label_name] = MinMaxScaler(feature_range=(0,1))
            # scalers[label_name] = StandardScaler()
            if bounds:
                scalers[label_name].fit(np.array(bounds[i]).reshape(-1,1))
            else:
                scalers[label_name].fit(data[:,i].reshape(-1,1))
            new_data[:,i] = scalers[label_name].transform(data[:,i].reshape(-1,1)).flatten()
        return scalers, new_data
    
    if label_scalers == None:
        label_scalers, labels = get_scalers(labels, labels_str)
    else:
        _, labels = get_scalers(labels, labels_str)
    
    if feature_scalers == None:
        feature_scalers, features = get_scalers(features, features_str,feature_bounds)
        # feature_scalers, features = get_scalers(features, features_str)
    else:
        _, features = get_scalers(features, features_str,feature_bounds)
        # _, features = get_scalers(features, features_str)

    for i in range(labels.shape[0]):
        for j in range(len(labels_str)):
            individuals[i].set_objective(labels_str[j],labels[i,j])

    for i in range(features.shape[0]):
        for j in range(len(features_str)):
            individuals[i].set_eval_parameter(features_str[j],features[i,j])

    return individuals, label_scalers, feature_scalers, labels_str, features_str

def inverse_transform_data(label_scalers:Dict[str,MinMaxScaler], feature_scalers:Dict[str,MinMaxScaler],individuals:List[Individual]) -> List[Individual]:
    """[summary]

    Args:
        label_scalers (Dict[str,MinMaxScaler]): [description]
        feature_scalers (Dict[str,MinMaxScaler]): [description]
        individuals (List[Individual]): [description]

    Returns:
        List[Individual]: [description]
    """
    individuals = [ind for ind in individuals if ind.IsFailed==False] # remove failed simulations

    labels = np.array([ind.objectives for ind in individuals])
    labels_str = [o.name for o in individuals[0].get_objectives_list()]

    features = np.array([ind.eval_parameters for ind in individuals])
    features_str = [f.name for f in individuals[0].get_eval_parameter_list()]
    
    for i in range(len(labels_str)):
        label_str = labels_str[i]
        labels[:,i] = label_scalers[label_str].inverse_transform(labels[:,i].reshape(-1,1)).flatten()
    
    for i in range(len(features_str)):
        feature_str = features_str[i]
        features[:,i] = feature_scalers[feature_str].inverse_transform(features[:,i].reshape(-1,1)).flatten()

    for i in range(labels.shape[0]):
        for j in range(len(labels_str)):
            individuals[i].set_objective(labels_str[j],labels[i,j])

    for i in range(features.shape[0]):
        for j in range(len(features_str)):
            individuals[i].set_eval_parameter(features_str[j],features[i,j])
    return individuals

def compute_mse(ml_individuals:List[Individual],eval_individuals:List[Individual]):
    passed_simulations = [i for i in range(len(eval_individuals)) if eval_individuals[i].IsFailed==False] # remove failed simulations

    ml_individuals = [ml_individuals[i] for i in passed_simulations]
    eval_individuals = [eval_individuals[i] for i in passed_simulations]

    objectives1 = np.array([ind.objectives for ind in ml_individuals])
    objectives2 = np.array([ind.objectives for ind in eval_individuals])
    return mean_squared_error(objectives1,objectives2)
    # mse = ((objectives1 - objectives2)**2).mean(axis=0)
    # return mse



"""
    Single objective differential evolution
"""
from random import shuffle,random
import operator
import subprocess, copy, math
import sys, os, shutil
from typing import List
from enum import Enum

import numpy as np
import matplotlib.pyplot as plt
from ..helpers import diversity, distance
from ..helpers import de_best_1_bin,de_rand_1_bin, mutation_parameters, de_mutation_type, simple,de_rand_1_bin_spawn,de_dmp, get_eval_param_matrix, get_objective_matrix, set_eval_parameters
from ..base import Parameter,Individual, Optimizer
from random import seed, gauss, random, randint
from tqdm import trange

class selection_type(Enum):
    best_design = 1
    pop_dist = 2

class SODE(Optimizer):
    def __init__(self,eval_command:str = "python evaluation.py", eval_folder:str = "Evaluation",pop_size:int=32, optimization_folder:str=None):
        super().__init__(name="SODE",eval_command=eval_command,eval_folder=eval_folder, opt_folder=optimization_folder)
        '''        
            Inputs:
                eval_script - Evaluation python script that will be called. Either Output.txt is read or an actual output is read. 
                eval_folder - folder to be copied into each individual evaluation directory. If this is null, the population directory isn't created and neither are the individual directories
                num_populations - number of populations to evaluate from the starting population
                pop_size - number of individuals in a given population
                optimization_folder - where optimization should start
        '''
        self.pop_size = pop_size
        self.individuals = None
        self.__mutation_params = mutation_parameters()

    # * Part of initialization    
    def add_eval_parameters(self,eval_params:List[Parameter]):
        self.eval_parameters = eval_params # Sets base class variable

    def add_objectives(self,objectives:List[Parameter]):
        self.objectives = objectives # Sets base class variable

    def add_performance_parameters(self,performance_params:List[Parameter] = None):
        self.performance_parameters = performance_params # Sets base class variable
    # *       

    # * Mutation Properties
    @property
    def mutation_params(self):
        return self.__mutation_params
    @mutation_params.setter
    def mutation_params(self,v):
        self.__mutation_params = v
    # * 

    def __set_eval_parameters__(self,y:np.ndarray):
        """
            only call this function within the class, do not expose to outside. once we have the parameters set, we might need to set the values based on an array.
        """
        parameters = copy.deepcopy(self.eval_parameters)
        for indx in range(len(parameters)):
            parameters[indx].value = y[indx]
        return parameters

    def start_doe(self,doe_individuals:List[Individual]=None, doe_size:int=128):
        """
            Starts a design of experiments. If the DOE has already started and there is an output file for an individual then the individual won't be evaluated 
        """        
        if doe_individuals is None:
            doe_individuals = []
            for i in trange(doe_size):
                parameters = copy.deepcopy(self.eval_parameters)
                for eval_param in parameters:
                    eval_param.value = np.random.uniform(eval_param.min_value,eval_param.max_value,1)[0]
                doe_individuals.append(Individual(eval_parameters=parameters,objectives=self.objectives, performance_parameters = self.performance_parameters))

        # * Begin the evaluation
        self.evaluate_population(individuals=doe_individuals,population_number=-1)
        # * Read the DOE
        individuals = self.read_population(population_number=-1)
        self.append_restart_file(individuals) # Create the restart file

    def optimize_from_population(self,pop_start:int,n_generations:int,sel_type:selection_type=selection_type.best_design):
        """
            Reads the values of a population, this can be a DOE or a previous evaluation
            Starts the optimization 

            Inputs:
                pop_start (-1 for DOE), reads the population folder and starts at pop_start+1
                n_generations - number of generations to run for
        """
        # * Read in all the results of the DOE, this should be done by a single thread
        # Check restart file, if not read the population
        
        individuals = self.read_restart_file()

        if (len(individuals)==0):
            individuals = self.read_population(population_number=pop_start)            
        if (len(individuals)<self.pop_size):
            raise Exception("Number of individuals in the restart file is less than the population size."
                + " lower the population size or increase the DOE count(if restarting from a DOE)")
        # Crossover and Mutate the doe individuals to generate the next individuals used in the population
        # Sort the population into [fill in here]
        self.load_history_file()
        individuals = sorted(individuals, key=operator.attrgetter('objectives'))
        individuals = individuals[:self.pop_size]
        if pop_start == -1: 
            pop_end = n_generations+pop_start
        else:
            pop_end = n_generations +pop_start+1

        for pop in trange(pop_start+1,pop_end):
            newIndividuals = self.__crossover_mutate__(individuals)
            self.evaluate_population(newIndividuals,pop) 
            newIndividuals = self.read_population(pop)
            
            pop_diversity = diversity(newIndividuals)       # Calculate diversity 
            pop_dist = distance(individuals,newIndividuals) # Calculate population distance
            if sel_type == selection_type.best_design:
                individuals.extend(newIndividuals)
                individuals = sorted(individuals, key=operator.attrgetter('objectives'))
                individuals = individuals[:self.pop_size]
            else:                
                individuals = self.select_individuals(individuals,newIndividuals)
                individuals = sorted(individuals, key=operator.attrgetter('objectives'))
            self.append_restart_file(individuals)
            self.append_history_file(pop,individuals[0],pop_diversity,pop_dist)
            
            if self.single_folder_eval:
                # Delete the population folder
                population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(pop_start))
                if os.path.isdir(population_folder):
                    shutil.rmtree(population_folder)
            pop_start+=1 # increment the population
    def __crossover_mutate__(self,individuals:List[Individual]):
        '''
            Applies Crossover and Mutate
        '''        
        nIndividuals = len(individuals)
        num_params = len(individuals[0].eval_parameters)        
        if self.mutation_params.mutation_type == de_mutation_type.de_best_1_bin:
            newIndividuals = de_best_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)

        elif self.mutation_params.mutation_type == de_mutation_type.de_rand_1_bin:
            shuffle(individuals)

            newIndividuals = de_rand_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)

        elif self.mutation_params.mutation_type == de_mutation_type.simple:
            shuffle(individuals)
            nCrossover = int(self.pop_size/2)

            nMutation = self.pop_size-nCrossover
            newIndividuals = simple(individuals=individuals,nCrossover=nCrossover,nMutation=nMutation,objectives=self.objectives,eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,mu=self.mutation_params.mu,sigma=self.mutation_params.sigma)

        elif self.mutation_params.mutation_type == de_mutation_type.de_rand_1_bin_spawn:
            shuffle(individuals)
            parents = individuals[0:self.mutation_params.nParents]
            
            newIndividuals = de_rand_1_bin_spawn(individuals=parents,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C,num_children=len(individuals))

        else:# self.mutation_params.mutation_type==mutation_parameters.de_dmp:
            newIndividuals = de_dmp(individuals=individuals,
                objectives=self.objectives, eval_parameters=self.eval_parameters, performance_parameters=self.performance_parameters)

        return newIndividuals
    
    def select_individuals(self,prevIndividuals:List[Individual],newIndividuals:List[Individual]):
        '''
            Select individuals using diversity and distance. Use this only for single objective type problems. This is not suitable for multi-objective. 

            Inputs:
                previndividuals - previous population
                newIndividuals - new population
            Citations:
                (described in) Ghosh, A., Das, S., Mallipeddi, R., Das, A. K., & Dash, S. S. (2017). A Modified Differential Evolution With Distance-based Selection for Continuous Optimization in Presence of Noise. IEEE Access, 5, 26944–26964. https://doi.org/10.1109/ACCESS.2017.2773825

                (Modified version of) S. Das, A. Konar, and U. K. Chakraborty, ‘‘Improved differential evolution algorithms for handling noisy optimization problems,’’ in Proc. IEEE Congr. Evol. Comput., vol. 2. Sep. 2005, pp. 1691–1698.
        '''
        F = get_objective_matrix(prevIndividuals)
        F_new = get_objective_matrix(newIndividuals)
        
        U,_,_ = get_eval_param_matrix(newIndividuals)
        X,_,_ = get_eval_param_matrix(prevIndividuals)
        
        deltaF = np.absolute(F-F_new)   # Compute the delta objective
        dist = np.sum(np.absolute(U-X),axis=1)         # Calculate distance

        individuals = list()
        for i in range(len(newIndividuals)):
            if F_new[i]/F[i] < 1: 
                individuals.append(newIndividuals[i])                                               # X_new[i,:] = U[i,:]
            elif (F_new[i]/F[i] > 1) and (random() <= math.exp(-deltaF[i]/dist[i])):
                individuals.append(newIndividuals[i])                                               # X_new[i,:] = U[i,:]
            else:
                individuals.append(prevIndividuals[i])                                              # X_new[i,:] = X[i,:]
        return individuals


from numpy import float32
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Linear, Module, Dropout
from typing import List
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

class SimpleLinearModel(Module):
    def __init__(self,in_channels:int,out_channels:int):
        """This class joins a bunch of linear layers together to predict the output size

        Args:
            in_channels (int): number of inputs channels
            out_channels (int): number of output channels
            h_sizes (List[int], optional): Any additional internal linear layers. Defaults to None.
        """        
        super(SimpleLinearModel, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.lin1 = nn.Linear(in_channels,64)
        self.lin2 = nn.Linear(64,128)
        self.lin3 = nn.Linear(128,128)
        self.lin4 = nn.Linear(128,out_channels)
    
    def forward(self,x):
              
        out1 = F.relu6(self.lin1(x))
        out2 = F.relu6(self.lin2(out1))
        out3 = F.relu6(self.lin3(out2))
        out4 = self.lin4(out3)
        return out4


class MultiLayerLinear(Module):
    """
    Model aims to predict both cp and y with spatial convolutions (MFConv).
    """
    # Multi-layer Perceptron 
    # https://discuss.pytorch.org/t/how-to-create-mlp-model-with-arbitrary-number-of-hidden-layers/13124/2

    # https://github.com/FrancescoSaverioZuppichini/Pytorch-how-and-when-to-use-Module-Sequential-ModuleList-and-ModuleDict

    def __init__(self,in_channels:int,out_channels:int,h_sizes:List[int]=None):
        """This class joins a bunch of linear layers together to predict the output size

        Args:
            in_channels (int): number of inputs channels
            out_channels (int): number of output channels
            h_sizes (List[int], optional): Any additional internal linear layers. Defaults to None.
        """        
        super(MultiLayerLinear, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.layers = nn.ModuleList()
        if (h_sizes!=None):
            self.layers.append(nn.Linear(in_channels,h_sizes[0],bias=True))
            self.layer_sizes = h_sizes
            for k in range(len(h_sizes)-1):
                self.layers.append(nn.Linear(h_sizes[k], h_sizes[k+1],bias=True))
            self.layers.append(nn.Linear(h_sizes[len(h_sizes)-1], out_channels,bias=True))
        else:
            self.layers.append(nn.Linear(in_channels,out_channels))
            self.layer_sizes = None
   
    def forward(self,x):
        out = x
        for i in range(len(self.layers)-1):
            out = F.relu6(self.layers[i](out))        
        return self.layers[-1](out) # dont activate the last layer

    def __str__(self):
        n_layers = len(self.layers)
        if (self.layer_sizes):
            layer_sizes = '-'.join([str(x) for x in self.layer_sizes])
            return f"MLL_IN-{self.in_channels}_OUT-{self.out_channels}_{layer_sizes}"
        else:
            return f"MLL_IN-{self.in_channels}_OUT-{self.out_channels}"

    def __repr__(self):
        return self.__str__()



import dataclasses_json
from ..base import Optimizer
from ..DOE import Experiment

def write_json(self,filename:str, doe_object:Experiment, opt_object:Optimizer):
    '''
        Writes the optimization setup to json
    '''
    assert len(doe_object.objectives) >0
    assert len(doe_object.eval_parameters)>0

    object_to_write = dict()
    objectives = list()
    for o in doe_object.objectives:
        objectives.append(o.perf_parameters.to_dict())
        
    eval_params = list()
    for o in doe_object.eval_parameters:
        eval_params.append(o.perf_parameters.to_dict())



from typing import List


def check_if_duplicates(listOfElems:List[str]) -> bool:
    """Checks if list has any duplicates

    Args:
        listOfElems (List[str]): List of strings 

    Returns:
        bool: True = duplicates present, False = no duplicates
    """
    
    if len(listOfElems) == len(set(listOfElems)):
        return False
    else:
        return True


import numpy as np
from ..base import Individual, Parameter
from typing import List
import math
from .mutate import get_eval_param_matrix, get_objective_matrix, set_eval_parameters
import random

def diversity(individuals:List[Individual]) -> float:
    """Computes the diversity of a population. Higher diversity value the better.

    Citations: 
        J. Hu, J. Zeng, and Y. Tan, ‘‘A diversity-guided particle swarm optimizer for dynamic environments,’’ in Bio-Inspired Computational Intelligence and Applications. Berlin, Germany: Springer, 2007, pp. 239–247

    Args:
        individuals (List[Individual]): List of individuals of a population 

    Returns:
        float: diversity value 
    """
    Np = len(individuals)
    D = len(individuals[0].eval_parameters)
    x_avg = 1/len(individuals) * np.sum(np.array([ind.eval_parameters for ind in individuals]),axis=0)
    
    L = np.zeros(Np)
    for i in range(Np):        
        x = individuals[i].eval_parameters
        for j in range(D):            
            L[i] += (x[j]-x_avg[j])**2
        L[i] = math.sqrt(L[i])
    
    pop_diversity = 1/(Np*np.max(L)) * np.sum(L)
    return pop_diversity

def distance(individuals:List[Individual], newIndividuals:List[Individual]):
    """Calculates the distance between individuals

    Args:
        individuals (List[Individual]): past list of individuals 
        newIndividuals (List[Individual]): updated list of individuals 

    Returns:
        float: distance
    """
    pop, _ , _ = get_eval_param_matrix(individuals)
    newPop, _ , _ = get_eval_param_matrix(newIndividuals)
    dist = np.sum(np.absolute(pop - newPop),axis=1)   # Compute distance
    return np.average(dist)




"""
    adjoint - gradient based optimization 
"""
import os, shutil
import copy, random
from typing import List, Tuple
from scipy.optimize import minimize, Bounds
import numpy as np
from tqdm import trange

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.optim import AdamW
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ..helpers import MultiLayerLinear 
from ..helpers import diversity, distance, set_eval_parameters
from ..helpers import non_dominated_sorting, find_intercepts, uniform_reference_points, sort_and_select_population
from ..helpers import compute_mse, transform_data
from ..base import Parameter, Individual, Optimizer
from .nsga3 import find_intercepts
individual_list = List[Individual]

@torch.no_grad()
def surrogate_objective_func(x0:np.ndarray,model:nn.Module,reference_points:List[np.ndarray],dist_index:int, labels_scaler:List[MinMaxScaler], features_scaler:List[MinMaxScaler]):
    """Objective function of adjoint using neural networks. The goal is to use this function to find values of x0 that minimize the distance to the reference point

    Args:
        x0 (np.ndarray): initial guess
        model (nn.Module): neural network model used for prediction
        reference_points (List[np.ndarray]): array of reference points along the pareto front
        intercepts (np.ndarray)

    Returns:
        (float): sum of all the fitnesses. This is the value to be minimized
    """
    x0_2 = copy.deepcopy(x0)
    for l in range(len(features_scaler)):
        x0_2[l] = features_scaler[l].transform(x0_2[l].reshape(-1,1))

    x0_2 = torch.as_tensor(x0_2,dtype=torch.float32)
    fitnesses = model(x0_2)
    fitnesses = fitnesses.detach().numpy()    
    for f in range(len(labels_scaler)):        
        fitnesses[f] = labels_scaler[f].inverse_transform(fitnesses[f].reshape(-1,1)*reference_points[dist_index][f])

    return np.sum(fitnesses[f])


class NSOPT(Optimizer):
    def __init__(self,eval_command:str = "python evaluation.py", eval_folder:str = "Evaluation", optimization_folder:str=None,single_folder_eval:bool=False, overwrite_input_file:bool=False, linear_network:List[int]=[64,64,64,64],epochs:int=200, train_test_split:float=0.8,pareto_resolution:int=32, min_method:str='Powell'):
        """NSOPT - Non-dominated Sorting Machine Learning Optimization Strategy for multi-objective applications.
            This strategy uses machine learning to build a surrogate model and scipy.optimizer.minimize to search for the values of $\vec{x}$ minimizes all objectives. The results will be a pareto front or surface or whatever that is in higher dimensions. 

        Args:
            eval_command (str, optional): This is the evaluation command that will be called inside the Individual's folder. Defaults to "python evaluation.py".
            eval_folder (str, optional): Folder that contains the executable files that are copied into each separate evaluation. Defaults to "Evaluation".
            optimization_folder (str, optional): Folder where the optimization and doe work should be stored in. Defaults to None.
            single_folder_eval (bool, optional): Evaluate within a single folder and not make a bunch of folders . Defaults to False.
            overwrite_input_file (bool, optional): whether or not to overwrite the input file with new data when restarting a simulation. Defaults to False.
            linear_network (list, optional): Size of MultiLinear network. Defaults to [64,64,64,64].
            epochs (int, optional): Number of epochs to train neural network for. Defaults to 200
            train_test_split (float, optional): Number of datapoints to be assigned to train and test. Defaults to 0.8.
            pareto_resolution (int, optional): indicates how many points to use to search for the pareto. This is also the population size. 
            min_method (str, optional): minimization method from https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html. Defaults to 'Powell'

        """
        super().__init__(name="NSOPT",eval_command=eval_command,eval_folder=eval_folder, opt_folder=optimization_folder,single_folder_eval=single_folder_eval,overwrite_input_file=overwrite_input_file)
        
        self.individuals = None
        self.linear_network = linear_network
        self.epochs = epochs
        self.train_test_split = train_test_split
        self.pareto_resolution = pareto_resolution
        self.model = None
        self.optimizer = None
        self.min_method = min_method

        self.label_scalers = None
        self.feature_scalers = None
        self.labels_str = None
        self.features_str = None
        self.n_inputs = None
        self.n_outputs = None


    def train(self,individuals:List[Individual],retrain:bool=False):
        """Trains the neural network to predict the output given an input 

        Args:
            individuals (List[Individual]): List of individuals to train the neural network
            retrain (bool, Optional): (True) retrains the existing model on new data. (False) create a new model for every population. Setting it to false means training takes longer but the model is more accurate because it renormalizes the inputs and outputs. 

        Returns:
            Tuple[float float]: Train Loss and test loss 
        """
        # * Normalizing the Data
        if retrain: # If we simply train it with more data then there's no reason create new normalization scalers
            normalized_individuals, label_scalers, feature_scalers, labels_str, features_str = transform_data(individuals,self.label_scalers,self.feature_scalers)
        else:
            normalized_individuals, label_scalers, feature_scalers, labels_str, features_str = transform_data(individuals)
        self.label_scalers = label_scalers
        self.feature_scalers = feature_scalers
        self.labels_str = labels_str
        self.features_str = features_str

        labels = torch.as_tensor(np.array([ind.objectives for ind in normalized_individuals]),dtype=torch.float32)
        features = torch.as_tensor(np.array([ind.eval_parameters for ind in normalized_individuals]),dtype=torch.float32)
        
        # Transform
        data = list(zip(features,labels))
        test_size = int(len(data)*(1-self.train_test_split))
        train_size = int(len(data) - test_size)
        train_dataset, test_dataset = train_test_split(data,test_size=test_size,train_size=train_size,shuffle=True)
        train_dl = DataLoader(train_dataset,batch_size=128,shuffle=True)
        test_dl = DataLoader(test_dataset,batch_size=128,shuffle=False)
        # * Defining the Model        
        n_inputs = features.shape[1]
        n_outputs = labels.shape[1]

        if self.model is None or retrain == False:
            self.model = MultiLayerLinear(n_inputs,n_outputs,h_sizes=self.linear_network)
            # self.optimizer = LBFGS(self.model.parameters(), lr=0.0001,history_size=100, max_eval=int(20*1.25), max_iter=20)
            self.optimizer = AdamW(self.model.parameters())
            self.n_inputs = n_inputs
            self.n_outputs = n_outputs 

        criterion = nn.MSELoss()
        
        for epoch in range(self.epochs):
            train_running_loss = 0 
            n_train = 0
            self.model.train()
            for i, (x, y) in enumerate(train_dl):
                batch_size = x.shape[0]
                self.optimizer.zero_grad()
                y_pred = self.model(x)
                loss = criterion(y_pred, y)
                loss.backward() # Zero gradients, backward pass, and update weights
                self.optimizer.step()
                # calculate the loss again for monitoring
                train_running_loss += loss.item()
                n_train += batch_size
            
            test_loss = 0
            n_test = 0
            self.model.eval()
            for j, (x,y) in enumerate(test_dl):
                batch_size = x.shape[0]
                y_pred = self.model(x)
                test_loss += criterion(y_pred, y).item()
                n_test += batch_size
            train_running_loss /= n_train
            test_loss /= n_test
        return train_running_loss, test_loss

    def optimize_from_population(self,pop_start:int,n_generations:int):
        """Starts the optimization by reading the values of a population. This can be a DOE or a previous evaluation 

        Args:
            pop_start (int): pop_start=-1 for DOE. Reads the population folder and starts at pop_start+1
            n_generations (int): Number of generations to iterate for 

        Raises:
            Exception: [description]
        """
        # * Read in all the results of the DOE, this should be done by a single thread
        # Check restart file, if not read the population
        self.load_history_file()
        ref_points = uniform_reference_points(len(self.objectives), p=self.pareto_resolution, scaling=None) # the -1 is there so that the number of reference points matches the number of individuals

        all_individuals = list() 

        newIndividuals = self.read_calculation_folder() # Use all individuals, there's no restart file
        newIndividuals = [item for sublist in newIndividuals for item in sublist]  # Flattens the list of lists
        lb = [ep.min_value for ep in self.eval_parameters]
        ub = [ep.max_value for ep in self.eval_parameters]
        bounds = Bounds(lb,ub)
        if (len(newIndividuals)<self.pareto_resolution):
            raise Exception("Number of individuals in the restart file is less than the population size."
                + " lower the population size or increase the DOE count(if restarting from a DOE)")

        # Do this before going into the train loop. This part of the code should happen after a new population is evaluated 
        individuals,best_point, worst_point, extreme_points = sort_and_select_population(individuals=newIndividuals,reference_points=ref_points, pop_size=self.pareto_resolution)
        
        all_individuals.extend(copy.deepcopy(newIndividuals))
        newIndividuals=individuals

        for pop in range(pop_start+1,pop_start+n_generations):  # Population Loop 
            ''' 
                Train a Neural network on Individuals. Initially this is all the individuals
            '''
            train_loss, test_loss = self.train(copy.deepcopy(all_individuals), False)   
            torch.save({'pop':pop,'state_dict':self.model.state_dict(),
                        'optimizer':self.optimizer.state_dict(),
                        'n_inputs':self.n_inputs,
                        'n_outputs':self.n_outputs},'ml_model.pt')   # This saved file can be used later when you want to plot for any input or output
            label_scalers = [self.label_scalers[l] for l in self.labels_str]
            feature_scalers = [self.feature_scalers[l] for l in self.features_str]
            '''
                Run parts of NSGA3 code to find intercepts 
            '''
            pareto_fronts = non_dominated_sorting(individuals,self.pareto_resolution)
            fitnesses = np.array([ind.objectives for f in pareto_fronts for ind in f])
            fitnesses *= -1            

            front_worst = np.max(fitnesses[:sum(len(f) for f in pareto_fronts),:],axis=0)
            intercepts = find_intercepts(extreme_points,best_point,worst_point,front_worst)

            features = np.array([ind.eval_parameters for ind in individuals])       # Normalized Features
            newIndividuals.clear()
            # surrogate_objective_func(res.x,self.model,ref_points,intercepts,o, best_point,label_scalers,feature_scalers)
            for o in range(self.pareto_resolution):
                x0 = features[random.randrange(0,len(features))]
                # least_squares(surrogate_objective_func,x0,jac=)
                res = minimize(surrogate_objective_func,x0,bounds=bounds,method=self.min_method,args=(self.model,ref_points,o,label_scalers,feature_scalers))
                surrogate_objective_func(res.x,self.model,ref_points,o,label_scalers,feature_scalers)                
                newIndividuals.append(
                    Individual(eval_parameters=set_eval_parameters(self.eval_parameters,copy.deepcopy(res.x)),
                        objectives=copy.deepcopy(self.objectives),performance_parameters=copy.deepcopy(self.performance_parameters))
                        )
            # newIndividuals = inverse_transform_data(self.label_scalers,self.feature_scalers,newIndividuals)
            # Evaluate
            self.evaluate_population(newIndividuals,pop)
            newIndividuals = self.read_population(pop)
            mse = compute_mse(individuals,newIndividuals)
            print(f"POP {pop} mse {mse:03e}")

            # Sort and select
            pop_diversity = diversity(newIndividuals)       # Calculate diversity 
            pop_dist = distance(individuals,newIndividuals) # Calculate population distance between past and future
            all_individuals.extend(copy.deepcopy(newIndividuals))
            newIndividuals.extend(individuals)              
            
            individuals,best_point, worst_point, extreme_points = sort_and_select_population(individuals=newIndividuals,reference_points=ref_points, pop_size=self.pareto_resolution)

            self.append_restart_file(newIndividuals)        # Keep the last designs
            self.append_history_file(pop,individuals[0],pop_diversity,pop_dist,train_loss,test_loss,mse)

            # if pop %4 ==0:
            #     all_individuals,_, _, _ = sort_and_select_population(individuals=all_individuals,reference_points=ref_points, pop_size=self.pareto_resolution*4)

            if self.single_folder_eval:
                # Delete the population folder
                population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(pop))
                if os.path.isdir(population_folder):
                    shutil.rmtree(population_folder)            

    def add_eval_parameters(self,eval_params:List[Parameter]):
        """Add evaluation parameters. This is part of the initialization    

        Args:
            eval_params (List[Parameter]): Add in a list of evaluation parameters 

        """
        self.eval_parameters = eval_params # Sets base class variable

    def add_objectives(self,objectives:List[Parameter]):
        """Add the objectives 

        Args:
            objectives (List[Parameter]): List of objectives 
        """
        self.objectives = objectives # Sets base class variable

    def add_performance_parameters(self,performance_params:List[Parameter] = None):
        """Add performance parameters 

        Args:
            performance_params (List[Parameter], optional): List of performance parameters. Defaults to None.
        """
        self.performance_parameters = performance_params # Sets base class variable 
    
    def start_doe(self,doe_individuals:List[Individual]=None,doe_size:int=128):
        """Starts a design of experiments. This generates the parameters for the individuals to be evaluated and executes each case. If the DOE has already started and there is an output file for an individual then the individual won't be evaluated    

        Args:
            doe_individuals (List[Individual], optional): List of individuals to evaluate. Defaults to None.
            doe_size (int, optional): Number of individuals to evaluate in the design of experiments. This is only used if doe_individuals is None. Defaults to 128.
        """
        if doe_individuals is None:
            doe_individuals = []
            for i in trange(doe_size):
                parameters = copy.deepcopy(self.eval_parameters)
                for eval_param in parameters:
                    eval_param.value = np.random.uniform(eval_param.min_value,eval_param.max_value,1)[0]
                doe_individuals.append(Individual(eval_parameters=parameters,objectives=self.objectives, performance_parameters = self.performance_parameters))
     
        # * Begin the evaluation
        self.evaluate_population(individuals=doe_individuals,population_number=-1)
        # * Read the DOE
        individuals = self.read_population(population_number=-1)
        self.append_restart_file(individuals)
        
        if self.single_folder_eval: 
            # Delete the population folder
            population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(-1))
            if os.path.isdir(population_folder):
                shutil.rmtree(population_folder)





from typing import List, Dict
import pandas as pd
import itertools
import copy
from ..base import Individual, Parameter
import numpy as np
from tqdm import trange
from doepy import build
from dataclasses_json import dataclass_json

class DOE:
    def __init__(self):
        """Initializes a design of experiments object
        """
        self.objectives = list()
        self.perf_parameters = list()
        self.eval_parameters = list()

        self.num_parameters = 0
        self.num_objectives = 0
        self.num_perf_parameter = 0

    def add_parameter(self,name:str = None, min_value:float = None ,max_value:float = None,value_if_failed:float = 100000, constr_less_than:float = None, constr_greater_than:float = None)->None:
        """Adds an evaluation parameter(x)  F(x) = objectives

        Args:
            name (str, optional): Name of the parameter example: x1. Defaults to None.
            min_value (float, optional): minimum value the parameter could be. Defaults to None.
            max_value (float, optional): maximum value the parameter could be. Defaults to None.
            value_if_failed (float, optional): on evaluation failure this value is used. Defaults to 100000.
            constr_less_than (float, optional): No need to input anything for this. Min and Max values are used to constrain. Defaults to None.
            constr_greater_than (float, optional): No need to input anything for this. Min and Max values are used to constrain. Defaults to None.
        """
        self.eval_parameters.append(Parameter(name, min_value,max_value,value_if_failed, constr_less_than, constr_greater_than))
        self.num_parameters = len(self.eval_parameters)
    
    def add_parameter_list(self,name:str=None,length:int=2,min_value:float=None ,max_value:float=None,value_if_failed:float=100000, constr_less_than:float = None, constr_greater_than:float = None)->None:
        for i in range(length):
            parameter_name = f"{name}_{i:d}"
            self.eval_parameters.append(Parameter(parameter_name, min_value,max_value,value_if_failed, constr_less_than, constr_greater_than))
        self.num_parameters = len(self.eval_parameters)

    def add_objectives(self,name:str = None, min_value:float = None ,max_value:float = None,value_if_failed:float = 100000, constr_less_than:float = None, constr_greater_than:float = None)->None:
        """Add an objective

        Args:
            name (str, optional): name of the objective. Defaults to None.
            min_value (float, optional): This you can leave blank. It doesn't do anything if you put something there. Defaults to None.
            max_value (float, optional): This you can leave blank. It doesn't do anything if you put something there. Defaults to None.
            value_if_failed (float, optional): On evaluation failure this value is used. Defaults to 100000.
            constr_less_than (float, optional): constraints the objective to being within the bounds. Defaults to None.
            constr_greater_than (float, optional): constraints the objective to being within the bounds. Defaults to None.
        """
        self.objectives.append(Parameter(name, min_value,max_value,value_if_failed, constr_less_than, constr_greater_than))
        self.num_objectives = len(self.objectives)

    def add_perf_parameter(self,name:str = None, min_value:float = None ,max_value:float = None,value_if_failed:float = 100000, constr_less_than:float = None, constr_greater_than:float = None)->None:
        """Add a performance parameter to keep track of

        Args:
            name (str, optional): Name of the parameter. Defaults to None.
            min_value (float, optional): [description]. Defaults to None.
            max_value (float, optional): [description]. Defaults to None.
            value_if_failed (float, optional): [description]. Defaults to 100000.
            constr_less_than (float, optional): [description]. Defaults to None.
            constr_greater_than (float, optional): [description]. Defaults to None.
        """
        self.perf_parameters.append(Parameter(name, min_value,max_value,value_if_failed, constr_less_than, constr_greater_than))
        self.num_perf_parameter = len(self.perf_parameters)

    
    def get_eval_values(self):
        """Based on the doe type, it generates the evaluation parameters for each individual and gets the values. 
            Typically not called directly. Use generate_doe.

        Returns:
            List[tuple]: list containaing an array of evaluation values. 
        """
        design =self.create_design()
        values= design[[self.eval_parameters[i].name for i in range(len(self.eval_parameters))]].values
        return [tuple(x) for x in values]

    def generate_doe(self) -> List[Individual]:
        """Generates the design of experiments. Returns a set of individuals with evaluation parameters all populated. 

        Returns:
            List[Individual]: individual designs
        """
        individuals=list()
        eval_values=self.get_eval_values()
        for i in trange(len(eval_values)):
            parameter = copy.deepcopy(self.eval_parameters)
            for indx in range(len(parameter)):
                parameter[indx].value=eval_values[i][indx]
            individuals.append(Individual(eval_parameters=parameter,objectives=self.objectives,performance_parameters=self.perf_parameters))
        return individuals
    
    def to_dict(self) -> Dict:
        """Export the doe settings to dictionary

        Returns:
            Dict: dictionary object with the class settings
        """
        settings = dict()
        settings['eval_parameters'] = [p.to_dict() for p in self.eval_parameters]
        settings['objectives'] = [o.to_dict() for o in self.objectives]
        settings['perf_parameters'] = [p.to_dict() for p in self.performance_parameters]
        settings['num_parameters'] = self.num_parameters
        settings['num_objectives'] = self.num_objectives
        settings['num_perf_parameter'] = self.num_perf_parameter
        return settings

    def from_dict(self,settings:dict):
        """Creates the class from a dictionary 

        Args:
            settings (dict): dictionary from to_dict
        """
        self.eval_parameters = [Parameter().from_dict(p) for p in settings['eval_parameters']]
        self.objectives = [Parameter().from_dict(p) for p in settings['objectives']]
        self.performance_parameters = [Parameter().from_dict(p) for p in settings['perf_parameters']]

        self.num_parameters = 0
        self.num_objectives = 0
        self.num_perf_parameter = 0

class Default(DOE):
    def __init__(self,number_of_evals:int):
        """Initializes a default doe. This doe defaults to randomizing the evaluation parameters from min to max

        Args:
            number_of_evals (int): [description]
        """
        super(Default, self).__init__()
        self.num_evals = number_of_evals
        self.eval_parameters=[]
        self.objectives=[]
        self.perf_parameters=[]

    def create_design(self):
        df = pd.DataFrame(data=[{self.eval_parameters[j].name : np.random.uniform(self.eval_parameters[j].min_value,self.eval_parameters[j].max_value,1)[0] for j in range(len(self.eval_parameters))} for i in range(self.num_evals) ])
        return df
    
    def to_dict(self) -> Dict:
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            Dict: dictionary object 
        """
        settings = DOE.to_dict(self) # call super class 
        settings['doe_name'] = 'default'
        settings['num_evals'] = self.num_evals

        return settings

    def from_dict(self, settings: dict):
        """creates the object from a dictionary 

        Args:
            settings (dict): dictionary object created by calling to_dict()
        """
        super().from_dict(settings)
        settings['num_evals'] = self.num_evals
        

class LatinHyperCube(DOE):   
    def __init__ (self,samples=10,levels=4):
        """This method is samples the design space in smaller cells.

        Args:
            samples (int, optional): number of evaluations . Defaults to 10.
            levels (int, optional): number of divisions of the evaluation parameters (this breaks it into cubes). Defaults to 4.

        Citations:
            https://pythonhosted.org/pyDOE/randomized.html
            https://doepy.readthedocs.io/en/latest/
        """
        super(LatinHyperCube, self).__init__()
        self.samples = samples 
        self.levels = levels

    def create_design(self) -> pd.DataFrame:
        """Creates a dataframe containing the doe design

        Returns:
            pd.DataFrame: dataframe object
        """
        param_dict = dict()
        for p in self.eval_parameters:
            r = np.linspace(p.min_value ,p.max_value ,self.levels)
            param_dict[p.name] = r.tolist()
        df = build.space_filling_lhs(param_dict, self.samples)
        return df
    
    def to_dict(self):
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            Dict: dictionary object 
        """
        settings = DOE.to_dict(self) # call super class 
        settings['doe_name'] = 'latinhypercube'
        settings['samples'] = self.samples
        settings['levels'] = self.levels
        return settings

    def from_dict(self, settings: dict):
        """creates the object from a dictionary 

        Args:
            settings (dict): dictionary object created by calling to_dict()
        """
        super().from_dict(settings)
        settings['samples'] = self.samples
        settings['levels'] = self.levels

class CCD(DOE):
    def __init__ (self,center_points:tuple=(4,4),alpha:str="o",face:str="ccc"):
        """Central Composite Design

        Args:
            center_points (tuple, optional): 4 - factorial block (4 = square) These are the edges. The other 4 is the + sign. You can change this. Defaults to (4,4).
            alpha (str, optional): "o" orthogonal or "r" for rotatable. Defaults to "o".
            face (str, optional): circumscribed "ccc" default, faced"ccf", inscribed "cci". Defaults to "ccc".
        """
        super(CCD, self).__init__()
        self.center= center_points
        self.alpha = alpha
        self.face = face 

    def create_design(self):
        """Creates a dataframe containing the doe design

        Returns:
            pd.DataFrame: dataframe object
        """
        param_dict = dict((p.name, [p.min_value, p.max_value]) for p in self.eval_parameters)    
        df = build.central_composite(param_dict,face=self.face)
        return df

    def to_dict(self):
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            Dict: dictionary object 
        """
        settings = DOE.to_dict(self) # call super class 
        settings['doe_name'] = 'ccd'
        settings['center_points'] = self.center_points
        settings['alpha'] = self.alpha
        settings['face'] = self.face
        return settings

    def from_dict(self, settings: dict):
        """creates the object from a dictionary 

        Args:
            settings (dict): dictionary object created by calling to_dict()
        """
        super().from_dict(settings)
        self.center_points = settings['center_points']
        self.alpha = settings['alpha']
        self.face =settings['face']

class BoxBehnken(DOE):
    def __init__(self,center_points:int=1):
        """Box Behnken design of experiments. Their primary advantage is in addressing the issue of where the experimental boundaries should be, and in particular to avoid treatment combinations that are extreme. Extreme = corner and star points. 

        Args:
            center (int, optional): Number of levels within the min and max values. Defaults to 4.
        """
        super(BoxBehnken, self).__init__()
        self.center_points=center_points
        
    def create_design(self):        
        param_dict = dict() 
        for p in self.eval_parameters:
            r = np.linspace(p.min_value ,p.max_value)
            param_dict[p.name] = r.tolist()
            
        df = build.box_behnken(param_dict,center=self.center_points)
        return df

    def to_dict(self):
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            Dict: dictionary object 
        """
        settings = DOE.to_dict(self) # call super class 
        settings['doe_name'] = 'fullfactorial'
        settings['levels'] = self.levels
        return settings

    def from_dict(self, settings: dict):
        """creates the object from a dictionary 

        Args:
            settings (dict): dictionary object created by calling to_dict()
        """
        super().from_dict(settings)
        self.levels = settings['levels']
    
class FullFactorial(DOE):
    def __init__(self,levels:int=4):
        """Factorial based design of experiments. number of evaluations scale with 2^(level-1)

        Args:
            levels (int, optional): Number of levels within the min and max values. Defaults to 4.
        """
        super(FullFactorial, self).__init__()
        self.levels=levels
        
    def create_design(self):        
        param_dict = dict() 
        for p in self.eval_parameters:
            r = np.linspace(p.min_value ,p.max_value ,self.levels)
            param_dict[p.name] = r.tolist()
            
        df = build.full_fact(param_dict)
        return df

    def to_dict(self):
        """Export the settings used to create the optimizer to dict. Also exports the optimization results if performed

        Returns:
            Dict: dictionary object 
        """
        settings = DOE.to_dict(self) # call super class 
        settings['doe_name'] = 'fullfactorial'
        settings['levels'] = self.levels
        return settings

    def from_dict(self, settings: dict):
        """creates the object from a dictionary 

        Args:
            settings (dict): dictionary object created by calling to_dict()
        """
        super().from_dict(settings)
        self.levels = settings['levels']
            

    

  





    










import numpy as np

def convert_to_ndarray(t) -> np.ndarray:
    """Converts a scalar or list to a numpy array

    Args:
        t (float,list): [description]

    Returns:
        np.ndarray: variable as an array
    """
    if type(t) is not np.ndarray and type(t) is not list: # Scalar
        t = np.array([t],dtype=float)
    elif (type(t) is list):
        t = np.array(t,dtype=float)
    return t


import torch
import numpy as np 

def gradient(y:torch.Tensor, x:torch.Tensor, grad_outputs=None):
        """Compute dy/dx @ grad_outputs"""
        if grad_outputs is None:
            grad_outputs = torch.ones_like(y)
        grad = torch.autograd.grad(y, [x], grad_outputs = grad_outputs, create_graph=True)[0]
        return grad

def jacobian(y:np.ndarray, x:np.ndarray) -> np.array:
    """Compute dy/dx = dy/dx @ grad_outputs
        for grad_outputs in [1, 0, ..., 0], [0, 1, 0, ..., 0], ...., [0, ..., 0, 1]

        Source: https://discuss.pytorch.org/t/how-to-compute-jacobian-matrix-in-pytorch/14968/12 

    Args:
        y (np.ndarray): Outputs
        x (np.ndarray): Inputs

    Returns:
        np.ndarray: Matrix [number of outputs, number of eval parameters]
    """
        
    y = torch.as_tensor(y)
    x = torch.as_tensor(x) 

    jac = torch.zeros(y.shape[0], x.shape[0]) 
    for i in range(y.shape[0]):
        grad_outputs = torch.zeros_like(y)
        grad_outputs[i] = 1
        jac[i] = gradient(y, x, grad_outputs = grad_outputs)
    return jac.cpu().detach().numpy()


from __future__ import absolute_import
from importlib import import_module     

from .copy_helper import copy
from .convert_to_ndarray import convert_to_ndarray
from .parallel_settings import parallel_settings
from .mutate import mutation_parameters, get_eval_param_matrix,get_objective_matrix, de_mutation_type, shuffle_population,de_best_1_bin,de_rand_1_bin,de_rand_1_bin_spawn,de_dmp,simple, set_eval_parameters
from .population_distance import distance, diversity
from .nsga_functions import non_dominated_sorting, find_extreme_points, find_intercepts, associate_to_niche, niching, uniform_reference_points, sort_and_select_population
from .post_processing import get_best, get_pop_best

if import_module('torch') is not None:
    from .jacobian import gradient, jacobian

from .MultiLayerLinear import MultiLayerLinear, SimpleLinearModel
from .list_functions import check_if_duplicates
from .nn_helpers import transform_data, inverse_transform_data, compute_mse, evaluation_func
from .read_input_file import read_input_to_dict


import errno
import os,shutil
 
def copy(src:str, dest:str,symlinks=False, ignore=None):
    """Copies a directory

    Args:
        src (str): source directory
        dest (str): destination directory
        symlinks (bool, optional): This parameter accepts True or False, depending on which the metadata of the original links or linked links will be copied to the new tree. Defaults to False.
        ignore ([type], optional): If ignore is given, it must be a callable that will receive as its arguments the directory being visited by copytree(), and a list of its contents, as returned by os.listdir(). Defaults to None.
    """
    for item in os.listdir(src):
        if '__pycache__' not in item:
            s = os.path.join(src, item)
            d = os.path.join(dest, item)
            try:
                if os.path.isdir(s):
                    shutil.copytree(s, d, symlinks, ignore)
                else:
                    shutil.copy2(s, d)
            except OSError as e:
                print('Directory not copied. Error: %s' % e)


import numpy as np
from collections import defaultdict
from typing import Tuple,List
from ..base import Individual
from itertools import chain

def non_dominated_sorting(individuals:List[Individual],k:int,first_front_only=False):
    """Loops through the list of individuals and sorts through them. 

    Citation: 
        Yuan, Y., Xu, H., & Wang, B. (2014). An Improved NSGA-III Procedure for Evolutionary Many-objective Optimization. Genetic and Evolutionary Computation Conference (GECCO 2014), 661–668. https://doi.org/10.1145/2576768.2598342

    Args:
        individuals (List[Individual]): all individuals of a population 
        k (int): number of individuals to select 
        first_front_only (bool, optional): gets only the best individuals. Defaults to False.

    Returns:
        List[Individual]: List containing fronts so List of lists of individuals. 
    """
    def dominates(x:np.ndarray,y:np.ndarray) -> bool:
        """Returns true if all or any the objectives of x are less than y

        Args:
            x (np.ndarray): a set of individual x's objective values 
            y (np.ndarray): a set of individual y's objective values 

        Returns:
            bool: True if individual x dominates individual y 
        """        
        b = np.all(x <= y) & np.any(x<y)
        return b

    map_fit_ind = defaultdict(list)
    for ind in individuals:
        map_fit_ind[ind].append(ind)
    fits = list(map_fit_ind.keys())

    current_front = []
    next_front = []
    dominating_fits = defaultdict(int)
    dominated_fits = defaultdict(list)

    # Rank first Pareto front
    for i, fit_i in enumerate(fits):
        for fit_j in fits[i+1:]:
            if dominates(fit_i.objectives,fit_j.objectives):
                dominating_fits[fit_j] += 1
                dominated_fits[fit_i].append(fit_j)
            elif dominates(fit_j.objectives,fit_i.objectives):
                dominating_fits[fit_i] += 1
                dominated_fits[fit_j].append(fit_i)
        if dominating_fits[fit_i] == 0:
            current_front.append(fit_i)

    fronts = [[]]
    for fit in current_front:
        fronts[-1].extend(map_fit_ind[fit])
    pareto_sorted = len(fronts[-1])

    # Rank the next front until all individuals are sorted or
    # the given number of individual are sorted.
    if not first_front_only:
        N = min(len(individuals), k)
        while pareto_sorted < N:
            fronts.append([])
            for fit_p in current_front:
                for fit_d in dominated_fits[fit_p]:
                    dominating_fits[fit_d] -= 1
                    if dominating_fits[fit_d] == 0:
                        next_front.append(fit_d)
                        pareto_sorted += len(map_fit_ind[fit_d])
                        fronts[-1].extend(map_fit_ind[fit_d])
            current_front = next_front
            next_front = []

    return fronts


def find_extreme_points(fitnesses, best_point, extreme_points=None):
    """Finds the individuals with extreme values for each objective function. These definitions need to be updated. I used to know all this. 

    Warning:
        Don't call this directly unless you know what you are doing

    Args:
        fitnesses ([ndarray]): how close individuals are to best point.1
        best_point ([ndarray]): array containing the best points
        extreme_points ([ndarray], optional): Points on the extreme of either objective 1 or objective 2. Defaults to None.

    Returns:
        ndarray: fitness of individuals. Description needs update

    """
    
    # Keep track of last generation extreme points
    if extreme_points is not None:
        fitnesses = np.concatenate((fitnesses, extreme_points), axis=0)

    # Translate objectives
    ft = fitnesses - best_point

    # Find achievement scalarizing function (asf)
    asf = np.eye(best_point.shape[0])
    asf[asf == 0] = 1e6
    asf = np.max(ft * asf[:, np.newaxis, :], axis=2)

    # Extreme point are the fitnesses with minimal asf
    min_asf_idx = np.argmin(asf, axis=1)
    return fitnesses[min_asf_idx, :]

def find_intercepts(extreme_points:List[np.ndarray], best_point:np.ndarray, current_worst:np.ndarray, front_worst:np.ndarray):
    """Find intercepts between the hyperplane defined by the extreme and best point and each axis with the ideal point as origin.

        Think of the hyperplane as a plane defined with one end being the best (minimum) and the other end being the extreme point (worst)
    
    Warning:
        Don't call this directly unless you know what you are doing

    Args:
        extreme_points (List[np.ndarray]): List of the extreme points [best, worst]. Example for 2 objectives [[16.34, - 4.07], [7.28, 6.69]]
        best_point (np.ndarray): best that minimizes all objectives. Example for 2 objectives [7.28, -4.07]
        current_worst (np.ndarray): worst individual that maximizes all objectives. Example for 2 objectives [16.34, 6.69]
        front_worst (np.ndarray): worst individual on pareto front. 

    Returns:
        np.ndarray: objective values that intercept the hyperplane axis. Example for 2 objectives [16.34, 6.69]
    """
    # Construct hyperplane sum(f_i^n) = 1
    b = np.ones(extreme_points.shape[1])    # np.ones is used to construct the axis    
    A = extreme_points - best_point         # Hyper plane with one end at best and the other at worst point 
    try:
        x = np.linalg.solve(A, b)           # Solves for where the plane intercepts the axis
    except np.linalg.LinAlgError:
        intercepts = current_worst
    else:
        intercepts = 1 / x

        if (not np.allclose(np.dot(A, x), b) or
                np.any(intercepts <= 1e-6) or
                np.any((intercepts + best_point) > current_worst)):
            intercepts = front_worst

    return intercepts


def associate_to_niche(fitnesses:np.ndarray, reference_points:np.ndarray, best_point:np.ndarray, intercepts:np.ndarray):
    """Associates individuals to reference points and calculates niche number. Corresponds to Algorithm 3 of Deb & Jain (2014).

    Args:
        fitnesses (np.ndarray): multi-dimensional array of unnormalized objective values from all individuals.
        reference_points (np.ndarray): multi-dimensional array representing the ideal normalized pareto front/surface. Example for 2 objectives [ [0,1], [0.25, 0.75],[0.5, 0.5], [0.75, 0.25], [1, 0] ]
        best_point (np.ndarray): Array representing the best individual's objective. Example for 2 objectives [7.28, -4.07]
        intercepts (np.ndarray): [description]

    Returns:
        [type]: [description]
    """
    # Normalize by ideal point and intercepts
    fn = (fitnesses - best_point) / (intercepts - best_point)

    # Create distance matrix
    fn = np.repeat(np.expand_dims(fn, axis=1), len(reference_points), axis=1)
    norm = np.linalg.norm(reference_points, axis=1)

    distances = np.sum(fn * reference_points, axis=2) / norm.reshape(1, -1)
    distances = distances[:, :, np.newaxis] * reference_points[np.newaxis, :, :] / norm[np.newaxis, :, np.newaxis]
    distances = np.linalg.norm(distances - fn, axis=2)

    # Retrieve min distance niche index
    niches = np.argmin(distances, axis=1)
    distances = distances[range(niches.shape[0]), niches]
    return niches, distances

def niching(individuals, k, niches, distances, niche_counts):
    """[summary]

    Args:
        individuals ([type]): [description]
        k ([type]): [description]
        niches ([type]): [description]
        distances ([type]): [description]
        niche_counts ([type]): [description]

    Returns:
        [type]: [description]
    """
    selected = []
    available = np.ones(len(individuals), dtype=np.bool_)
    while len(selected) < k:
        # Maximum number of individuals (niches) to select in that round
        n = k - len(selected)

        # Find the available niches and the minimum niche count in them
        available_niches = np.zeros(len(niche_counts), dtype=np.bool_)
        available_niches[np.unique(niches[available])] = True
        min_count = np.min(niche_counts[available_niches])

        # Select at most n niches with the minimum count
        selected_niches = np.flatnonzero(np.logical_and(available_niches, niche_counts == min_count))
        np.random.shuffle(selected_niches)
        selected_niches = selected_niches[:n]

        for niche in selected_niches:
            # Select from available individuals in niche
            niche_individuals = np.flatnonzero(np.logical_and(niches == niche, available))
            np.random.shuffle(niche_individuals)

            # If no individual in that niche, select the closest to reference
            # Else select randomly
            if niche_counts[niche] == 0:
                sel_index = niche_individuals[np.argmin(distances[niche_individuals])]
            else:
                sel_index = niche_individuals[0]

            # Update availability, counts and selection
            available[sel_index] = False
            niche_counts[niche] += 1
            selected.append(individuals[sel_index])
    return selected


def uniform_reference_points(nobj, p=4, scaling=None):
    """Generate reference points uniformly on the hyperplane intersecting each axis at 1. The scaling factor is used to combine multiple layers of reference points.

    Args:
        nobj (int): number of objectives
        p (int, optional): number of references per objective. Defaults to 4.
        scaling ([type], optional): [description]. Defaults to None.
    
    Returns:
        List[Tuple(float)]: List of normalized objective function values representing the reference minimum values of each objective. Think of it as a reference pareto front/surface/(whatever in 3+ dimensions )
        
    """
    def gen_refs_recursive(ref, nobj, left, total, depth):
        points = []
        if depth == nobj - 1:
            ref[depth] = left / total
            points.append(ref)
        else:
            for i in range(left + 1):
                ref[depth] = i / total
                points.extend(gen_refs_recursive(ref.copy(), nobj, left - i, total, depth + 1))
        return points

    ref_points = np.array(gen_refs_recursive(np.zeros(nobj), nobj, p, p, 0))
    if scaling is not None:
        ref_points *= scaling
        ref_points += (1 - scaling) / nobj

    return ref_points

def sort_and_select_population(individuals:List[Individual], reference_points:np.ndarray,pop_size:int):
        """Takes a list of individuals, finds the fronts and the best designs
            
            Code is a combination from deap and yarpiz
            https://github.com/DEAP
            https://yarpiz.com/456/ypea126-nsga3

        Args:
            individuals (List[Individual]): List of individuals from a DOE or POP or anything really
            reference_points (np.ndarray): reference points along the pareto front.  
            pop_size (int): Population size to consider 

        Raises:
            Exception: something bad has happened

        Returns:
            (tuple): containing

                **chosen** (List[List[int]]): individuals along front in order of best to worst front 
                **best_point** (int): index of best individual 
                **worst_point** (int): index of worst individual 
                **extreme_points** (List[int]):  indexies of the extreme point 
        """
        
        if (pop_size>len(individuals)):
            raise Exception("population size needs to be <= the number of individuals")
        

        pareto_fronts = non_dominated_sorting(individuals,pop_size)
        fitnesses = np.array([ind.objectives for f in pareto_fronts for ind in f])
        fitnesses *= -1

        best_point = np.min(fitnesses,axis=0)
        worst_point = np.max(fitnesses,axis=0)

        extreme_points = find_extreme_points(fitnesses,best_point)
        front_worst = np.max(fitnesses[:sum(len(f) for f in pareto_fronts),:],axis=0)
        intercepts = find_intercepts(extreme_points,best_point,worst_point,front_worst)
        niches, dist = associate_to_niche(fitnesses, reference_points, best_point, intercepts)
        
        # Get counts per niche for individuals in all front but the last
        niche_counts = np.zeros(len(reference_points), dtype=int)
        index, counts = np.unique(niches[:-len(pareto_fronts[-1])], return_counts=True)
        niche_counts[index] = counts

        # Choose individuals from all fronts but the last
        chosen = list(chain(*pareto_fronts[:-1]))

        # Use niching to select the remaining individuals
        sel_count = len(chosen)
        n = pop_size - sel_count
        selected = niching(pareto_fronts[-1], n, niches[sel_count:], dist[sel_count:], niche_counts)
        chosen.extend(selected)

        return chosen, best_point, worst_point, extreme_points



from __future__ import absolute_import
from importlib import import_module     

from .nsga3 import NSGA3
from .sode import SODE
from .nsopt import NSOPT

if import_module('torch') is not None:
    from .nsga3_ml import NSGA3_ML
else:
    print("torch is not installed, skilling NSGA3_ML")



from glennopt.base import Parameter
import re

def read_input_to_dict(input_file:str="input.dat"):
    x = dict()
    with open(input_file, "r") as f: 
        parameter_name = ""
        for line in f:
            split_val = line.split('=')
            if len(split_val)==2: # x1 = 2 # Grab the 2
                matches = re.findall("(\w+)_(\d+$)",split_val[0])
                if len(matches)>0:
                    parameter_name = matches[0][0]
                    if parameter_name not in x.keys():
                        x[parameter_name] = []
                    x[parameter_name].append(float(split_val[1]))
                else:
                    parameter_name = split_val[0]
                    x[parameter_name] = float(split_val[1])

    return x

    



from dataclasses import dataclass, field
from enum import Enum
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class parallel_settings:
    """These settings control how the optimizer will execute. 
    Example: 
        .. code-block: python
            config = parallel_settings()
            ns = NSGA3()
            ns.parallel_settings = config
    
    Args:
        concurrent_executions (int): Number of concurrent executions per machine 
        cores_per_execution (int): number of cores per exectuion. Defaults to 1. Set it to 0 to ignore using cores per execution and just go with concurrent executions 
        execution_timeout (int): execution timeout in minutes. Defaults to 10 minutes 
        machine_filename (str): path to the machine file. Defaults to 'machinefile.txt'
        database_filename (str): path to the database. Defaults to 'database.csv'
        ignore_cores (bool): indicate whether to ignore the number of cores per execution and just execute the code

    """
    concurrent_executions: int = field(repr=True,default=1)
    cores_per_execution: int = field(repr=True,default=1)
    execution_timeout: int = field(repr=True,default=10)
    machine_filename = 'machinefile.txt'
    database_filename = 'database.csv'



"""
    NSGA3 - Non-dominated sorting genetic algorithm
"""
import os, shutil
import subprocess, copy, math
from random import seed, gauss,random,uniform,randint
from typing import List
from dataclasses import dataclass, field
import numpy as np 
import glob
from tqdm import trange
from ..helpers import diversity, distance
from ..helpers import non_dominated_sorting, find_extreme_points, find_intercepts, associate_to_niche, niching, uniform_reference_points, sort_and_select_population
from ..base import Parameter, Individual, Optimizer
from ..helpers import de_best_1_bin,de_rand_1_bin, mutation_parameters, de_mutation_type, simple,de_rand_1_bin_spawn,de_dmp, get_eval_param_matrix, get_objective_matrix, set_eval_parameters

individual_list = List[Individual]

class NSGA3(Optimizer):
    def __init__(self,eval_command:str = "python evaluation.py", eval_folder:str = "Evaluation",pop_size:int=128, optimization_folder:str=None,single_folder_eval=False, overwrite_input_file=False, pareto_resolution:int=4):
        """
            NSGA-3 multi-dimensional optimizer. This version has been tweaked to include restart capabilities. It can also keep track of additional parameters that can be considered part of the constraints.

            Each evaluation can occur in a separate folder (simulations) or without folders (analytical)
            https://www.egr.msu.edu/~kdeb/papers/k2012009.pdf

        Args:
            eval_command (str, optional): Command that will be executed in the evaluation folder. Defaults to "python evaluation.py".
            eval_folder (str, optional): folder to be copied into each individual evaluation directory. If this is null, the population directory isn't created and neither are the individual directories. Defaults to "Evaluation".
            pop_size (int, optional): number of populations to evaluate from the starting population. Defaults to 128.
            optimization_folder (str, optional): Folder where the optimization and doe work should be stored in. Defaults to None.
            single_folder_eval (bool, optional): where optimization should start. Defaults to False.
            overwrite_input_file(bool, optional): Specifies whether or not to overwrite the input file when restarting a simulation. Defaults to False.
            pareto_resolution (int, optional): This specifies the number of reference points on the pareto front, 4 is good for 2 objectives. This number should increase if more objectives are needed. 2^n objectives something like that would work. 

        """
        super().__init__(name="nsga3",eval_command=eval_command,eval_folder=eval_folder, opt_folder=optimization_folder,single_folder_eval=single_folder_eval,overwrite_input_file=overwrite_input_file)
        
        self.pop_size = pop_size
        self.individuals = None 
        self.pareto_resolution = pareto_resolution       
        self.__mutation_params = mutation_parameters()
        
        
    def add_eval_parameters(self,eval_params:List[Parameter]):
        """Add evaluation parameters. This is part of the initialization    

        Args:
            eval_params (List[Parameter]): Add in a list of evaluation parameters 

        """
        self.eval_parameters = eval_params # Sets base class variable

    def add_objectives(self,objectives:List[Parameter]):
        """Add the objectives 

        Args:
            objectives (List[Parameter]): [description]
        """
        self.objectives = objectives # Sets base class variable

    def add_performance_parameters(self,performance_params:List[Parameter] = None):
        """Add performance parameters 

        Args:
            performance_params (List[Parameter], optional): [description]. Defaults to None.
        """
        self.performance_parameters = performance_params # Sets base class variable
      
    
    # * Mutation Properties
    @property
    def mutation_params(self):
        """Get Mutation parameters

        Returns:
            mutation_parameters: parameter class describes the mutation
        """
        return self.__mutation_params

    @mutation_params.setter
    def mutation_params(self,v:mutation_parameters):
        """Setter for mutation parameters 

        Args:
            v (mutation_parameters): class describing the mutation parameter s
        """
        self.__mutation_params = v
    # * 

    def start_doe(self,doe_individuals:List[Individual]=None,doe_size:int=128):
        """Starts a design of experiments. This generates the parameters for the individuals to be evaluated and executes each case. If the DOE has already started and there is an output file for an individual then the individual won't be evaluated    

        Args:
            doe_individuals (List[Individual], optional): List of individuals. Defaults to None.
            doe_size (int, optional): [description]. Defaults to 128.
        """
        if doe_individuals is None:
            doe_individuals = []
            for i in trange(doe_size):
                parameters = copy.deepcopy(self.eval_parameters)
                for eval_param in parameters:
                    eval_param.value = np.random.uniform(eval_param.min_value,eval_param.max_value,1)[0]
                doe_individuals.append(Individual(eval_parameters=parameters,objectives=self.objectives, performance_parameters = self.performance_parameters))
     
        # * Begin the evaluation
        self.evaluate_population(individuals=doe_individuals,population_number=-1)
        # * Read the DOE
        individuals = self.read_population(population_number=-1)
        self.append_restart_file(individuals)
        
        if self.single_folder_eval: 
            # Delete the population folder
            population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(-1))
            if os.path.isdir(population_folder):
                shutil.rmtree(population_folder)

       

    def optimize_from_population(self,pop_start:int,n_generations:int):
        """Reads the values of a population, this can be a DOE or a previous evaluation
            Starts the optimization 

            Inputs:
                pop_start (-1 for DOE), reads the population folder and starts at pop_start+1
                n_generations - number of generations to run for
        """
        # * Read in all the results of the DOE, this should be done by a single thread
        # Check restart file, if not read the population
        self.load_history_file()
        individuals = self.read_restart_file()

        if (len(individuals)==0):                               
            individuals = self.read_calculation_folder()

        if (len(individuals)<self.pop_size):
            raise Exception("Number of individuals in the restart file is less than the population size."
                + " lower the population size or increase the DOE count(if restarting from a DOE)")

        # Crossover and Mutate the doe individuals to generate the next individuals used in the population
        # Sort the population into [fill in here]
        ref_points = uniform_reference_points(len(self.objectives), p=self.pareto_resolution, scaling=None)
        individuals,best_point, worst_point, extreme_points = sort_and_select_population(individuals=individuals,reference_points=ref_points, pop_size=self.pop_size)
        self.__optimize__(individuals=individuals,n_generations=n_generations,pop_start=pop_start+1, reference_points=ref_points)


    def __optimize__(self,individuals:individual_list,n_generations:int,pop_start:int, reference_points:np.ndarray):
        """ NSGA-III main loop
            Note: This function will read given starting population's results in, perform necessary crossover and mutation to generate enough individuals for the next iteration (self.pop_size)

            Inputs:
                individuals - list of individuals to evaluate
                n_generations - number of generations to loop through
                pop_start - starting population number
        """
        
        nIndividuals = len(individuals)
         
        # * Loop through all individuals
        for pop in range(pop_start,pop_start+n_generations):
            newIndividuals = self.__crossover_mutate__(individuals)

            # Evaluate
            self.evaluate_population(newIndividuals,pop_start)            
            newIndividuals = self.read_population(pop_start)
            # Sort and select
            pop_diversity = diversity(newIndividuals)       # Calculate diversity 
            pop_dist = distance(individuals,newIndividuals) # Calculate population distance between past and future
            newIndividuals.extend(individuals) # add the previous population to the pool                                    
            individuals,best_point, worst_point, extreme_points = sort_and_select_population(newIndividuals,reference_points, self.pop_size)            
            self.append_restart_file(individuals)        # Keep the last designs
            
            self.append_history_file(pop,individuals[0],pop_diversity,pop_dist)

            if self.single_folder_eval:
                # Delete the population folder
                population_folder = os.path.join(self.optimization_folder,self.__check_population_folder__(pop_start))
                if os.path.isdir(population_folder):
                    shutil.rmtree(population_folder)
            pop_start+=1 # increment the population
        # * End Loop through all individuals
    
    
    
    

    def __crossover_mutate__(self,individuals:List[Individual]):
        """[summary]

        Args:
            individuals (List[Individual]): [description]

        Returns:
            [type]: [description]
        """

        nIndividuals = len(individuals)
        num_params = len(individuals[0].eval_parameters)        
        if self.mutation_params.mutation_type == de_mutation_type.de_best_1_bin:
            newIndividuals = de_best_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)
        elif self.mutation_params.mutation_type == de_mutation_type.de_rand_1_bin:
            newIndividuals = de_rand_1_bin(individuals=individuals,objectives=self.objectives,
                eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,
                F=self.mutation_params.F,C=self.mutation_params.C)
        elif self.mutation_params.mutation_type == de_mutation_type.de_dmp:
            newIndividuals = de_dmp(individuals=individuals,
                objectives=self.objectives, eval_parameters=self.eval_parameters, performance_parameters=self.performance_parameters)
        else:               # self.mutation_params.mutation_type == de_mutation_type.simple
            nCrossover =  int(self.pop_size/2)
            nMutation = self.pop_size-nCrossover
            newIndividuals = simple(individuals=individuals,nCrossover=nCrossover,nMutation=nMutation,objectives=self.objectives,eval_parameters=self.eval_parameters,performance_parameters=self.performance_parameters,mu=self.mutation_params.mu,sigma=self.mutation_params.sigma)

        return newIndividuals
    
    
    
    




import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
from tqdm import trange
from typing import List

from ..base import Individual
from .convert_to_ndarray import convert_to_ndarray
from .nsga_functions import non_dominated_sorting

def get_best(individuals:List[Individual],pop_size:int):
    """Gets the best individual vs Pop. Some populations won't generate a better design but the best design will always be carried to the next population for crossover + mutation     
       
        Important: Call this function with inputs from ``individuals = ns.read_calculation_folder()``

    Args:
        individuals (List[Individual]): Takes in a list of the individuals
        pop_size (int): population size i.e. how many individuals

    Returns:
        (tuple): tuple containing:

            **objectives** (List[Parameter]): numpy array of best objective values for each population. For multi-objective problems use best fronts for better representation of design space

            **pop_folders** (List[int]): this is a list of populations 

            **best_fronts** (List[List[Individual]]): List of individuals contained in best fronts, empty list if single objective
    """

    best_individuals,best_fronts = get_pop_best(individuals)

    objectives = list()
    nobjectives = len(individuals[0][0].objectives)
    keys = list(best_individuals.keys())
    for i in range(len(individuals)):
        key = keys[i]
        temp_objectives = list()
        if i == 0:
            for o in range(nobjectives):
                temp_objectives.append(best_individuals[key][o].objectives[o])
        else:
            for o in range(nobjectives):
                if best_individuals[key][o].objectives[o] < objectives[-1][o]:
                    temp_objectives.append(best_individuals[key][o].objectives[o])
                else:
                    temp_objectives.append(objectives[-1][o])
        objectives.append(temp_objectives)
    
    # Read calculation folder
    best_fronts = list()
    if nobjectives>1:
        rolling_best = list()
        for i in trange(len(individuals), desc='Running Non-dimensional sorting'):
            pop_individuals = individuals[i]
            rolling_best.extend(pop_individuals)
            best_fronts.append(non_dominated_sorting(rolling_best,len(pop_individuals),True))
            if (len(rolling_best)>pop_size):                
                rolling_best = rolling_best[-pop_size:] # Keep only the pop_size
        
    pop_folders = keys
    return convert_to_ndarray(objectives), pop_folders, best_fronts  # First front

def get_pop_best(individuals:List[Individual]):
    '''
        Gets the best individuals from each population (not rolling best)
        typically you would use opt.read_calculation_folder() where opt is an object representing your nsga3 or sode class.
        
        Note: Important: Call this function with inputs from `individuals = ns.read_calculation_folder()`

        Returns:
            (tuple): tuple containing:

            **best_individuals** (List[Dict[str,List[Individual]]]): this is an array of individuals that are best at each objective

                |    [ 
                |        POP001: [best_individual_objective1, best_individual,objective2, best_individual,objective3], best_individual_compromise
                |        POP002: [best_individual_objective1, best_individual,objective2, best_individual,objective3], best_individual_compromise
                |        POP003: [best_individual_objective1, best_individual,objective2, best_individual,objective3], best_individual_compromise
                |    ]

            **comp_individuals** (List[Individual]): this is an array of individuals that is the best compromise between all the objectives

    '''
    # Read calculation folder
    
    # (Compromise Target) At the minimum index of each objective what are the values of the other objectives
    nobjectives = len(individuals[0][0].objectives)
    best_fronts = list()

    best_individuals = dict()

    for pop_individuals in individuals:
        pop = pop_individuals[0].population
        if nobjectives>1:
            best_fronts.append(non_dominated_sorting(pop_individuals,len(pop_individuals),True))
        
        for ind in pop_individuals:
            if pop not in best_individuals.keys():        # Prepopulate
                best_individuals[pop] = list()
                for o in range(nobjectives):
                    best_individuals[pop].append(ind) 
            else:                                       # Compare   
                for o in range(nobjectives): # Checks for the best objective
                    current_best = best_individuals[pop][o].objectives[o]
                    if ind.objectives[o]<current_best:
                        best_individuals[pop][o] = ind
    
    return best_individuals, best_fronts

def plot_pareto(best_fronts,pop,objective1_index,objective2_index):
    '''

    '''
    fig,ax = plt.subplots()

    colors = cm.rainbow(np.linspace(0, 1, len(best_fronts)))        
    indx = 0
    legend_labels = []
    # Scan the pandas file, grab objectives for each population
    for ind_list in best_fronts:
        obj1_data = []
        obj2_data = []
        c=colors[indx]
        for ind in ind_list:
            obj1_data.append(ind.objectives[objective1_index])
            obj2_data.append(ind.objectives[objective2_index])
        # Plot the gathered data
        ax.scatter(obj1_data, obj2_data, color=c, s=10,alpha=0.5)
        legend_labels.append(pop[indx])
        indx+=1

    ax.set_xlabel(obj1_name)
    ax.set_ylabel(obj2_name)
    if xlim is not None:
        ax.set_xlim(xlim[0],xlim[1])
    if ylim is not None:
        ax.set_ylim(ylim[0],ylim[1])
    ax.legend(legend_labels)
    fig.canvas.draw()
    fig.canvas.flush_events()
    plt.show()
